import numpy as np
from scipy.interpolate import interp1d
import tomviz.operators


class AutoTiltAxisShiftAlignmentOperator(tomviz.operators.CancelableOperator):

    def transform(self, dataset):
        """Automatic align the tilt axis to the center of images"""
        self.progress.maximum = 1

        # Get Tilt angles
        tilt_angles = dataset.tilt_angles

        tiltSeries = dataset.active_scalars
        if tiltSeries is None:
            raise RuntimeError("No scalars found!")

        Nx, Ny, Nz = tiltSeries.shape

        shifts = (np.linspace(-20, 20, 41)).astype('int')
        numberOfSlices = 5  # number of slices used for recon

        # randomly choose slices with top 50% total intensities
        tiltSeriesSum = np.sum(tiltSeries, axis=(1, 2))
        temp = tiltSeriesSum.argsort()[Nx // 2:]
        slices = temp[np.random.permutation(temp.size)[:numberOfSlices]]
        print('Reconstruction slices:')
        print(slices)

        I = np.zeros(shifts.size)

        self.progress.maximum = shifts.size - 1
        step = 1

        for i in range(shifts.size):
            if self.canceled:
                return
            shiftedTiltSeries = np.roll(
                tiltSeries[slices, :, :, ], shifts[i], axis=1)
            for s in range(numberOfSlices):
                self.progress.message = ('Reconstructing slice No.%d/%d with a '
                                         'shift of %d pixels' %
                                         (s, numberOfSlices, shifts[i]))

                recon = wbp2(shiftedTiltSeries[s, :, :],
                             tilt_angles, Ny, 'ramp', 'linear')
                I[i] = I[i] + np.amax(recon)

            step += 1
            self.progress.value = step

        print('shift: %d' % shifts[np.argmax(I)])

        result = np.roll(tiltSeries, shifts[np.argmax(I)], axis=1)
        result = np.asfortranarray(result)

        # Set the result as the new scalars.
        dataset.active_scalars = result


def wbp2(sinogram, angles, N=None, filter="ramp", interp="linear"):
    if sinogram.ndim != 2:
        raise ValueError('Sinogram must be 2D')
    (Nray, Nproj) = sinogram.shape
    if Nproj != angles.size:
        raise ValueError('Sinogram does not match angles!')

    interpolation_methods = ('linear', 'nearest', 'spline', 'cubic')
    if interp not in interpolation_methods:
        raise ValueError("Unknown interpolation: %s" % interp)
    if not N:  # if output size is not given
        N = int(np.floor(np.sqrt(Nray**2 / 2.0)))

    ang = np.double(angles) * np.pi / 180.0
    # Create Fourier filter
    F = makeFilter(Nray, filter)
    # Pad sinogram for filtering
    s = np.lib.pad(sinogram, ((0, F.size - Nray), (0, 0)),
                   'constant', constant_values=(0, 0))
    # Apply Fourier filter
    s = np.fft.fft(s, axis=0) * F
    s = np.real(np.fft.ifft(s, axis=0))
    # Change back to original
    s = s[:Nray, :]

    # Back projection
    recon = np.zeros((N, N))
    center_proj = Nray // 2  # Index of center of projection
    [X, Y] = np.mgrid[0:N, 0:N]
    xpr = X - int(N) // 2
    ypr = Y - int(N) // 2

    for j in range(Nproj):
        t = ypr * np.cos(ang[j]) - xpr * np.sin(ang[j])
        x = np.arange(Nray) - center_proj
        if interp == 'linear':
            bp = np.interp(t, x, s[:, j], left=0, right=0)
        elif interp == 'spline':
            interpolant = interp1d(
                x, s[:, j], kind='slinear', bounds_error=False, fill_value=0)
            bp = interpolant(t)
        else:
            interpolant = interp1d(
                x, s[:, j], kind=interp, bounds_error=False, fill_value=0)
            bp = interpolant(t)
        recon = recon + bp

    # Normalize
    recon = recon * np.pi / 2 / Nproj
    return recon

# Filter (1D) projections.


def makeFilter(Nray, filterMethod="ramp"):
    # Calculate next power of 2
    N2 = 2**np.ceil(np.log2(Nray))
    # Make a ramp filter.
    freq = np.fft.fftfreq(int(N2)).reshape(-1, 1)
    omega = 2 * np.pi * freq
    filter = 2 * np.abs(freq)

    if filterMethod == "ramp":
        pass
    elif filterMethod == "shepp-logan":
        filter[1:] = filter[1:] * np.sin(omega[1:]) / omega[1:]
    elif filterMethod == "cosine":
        filter[1:] = filter[1:] * np.cos(filter[1:])
    elif filterMethod == "hamming":
        filter[1:] = filter[1:] * (0.54 + 0.46 * np.cos(omega[1:] / 2))
    elif filterMethod == "hann":
        filter[1:] = filter[1:] * (1 + np.cos(omega[1:] / 2)) / 2
    elif filterMethod == "none":
        filter[:] = 1
    else:
        raise ValueError("Unknown filter: %s" % filterMethod)

    return filter



def transform(dataset):
    """Set negative voxels to zero"""

    data = dataset.active_scalars

    data[data < 0] = 0 # Set negative voxels to zero

    # Set the result as the new scalars.
    dataset.active_scalars = data



# -*- coding: utf-8 -*-

###############################################################################
# This source file is part of the Tomviz project, https://tomviz.org/.
# It is released under the 3-Clause BSD License, see "LICENSE".
###############################################################################


class Progress(object):
    """
    Class used to update operator progress.
    """

    def __init__(self, operator):
        self._operator = operator

    @property
    def maximum(self):
        """
        Property defining the maximum progress value
        """
        return self._operator._operator_wrapper.progress_maximum

    @maximum.setter
    def maximum(self, value):
        self._operator._operator_wrapper.progress_maximum = value

    @property
    def value(self):
        """
        Property defining the current progress value
        """
        return self._operator._operator_wrapper.progress_value

    @value.setter
    def value(self, value):
        """
        Updates the progress of the the operator.

        :param value The current progress value.
        :type value: int
        """
        self._operator._operator_wrapper.progress_value = value

    @property
    def message(self):
        """
        Property defining the current progress message
        """
        return self._operator._operator_wrapper.progress_message

    @message.setter
    def message(self, msg):
        """
        Updates the progress message of the the operator.

        :param msg The current progress message.
        :type msg: str
        """
        self._operator._operator_wrapper.progress_message = msg

    def _data(self, value):
        # Make sure this is a vtkDataObject before setting
        from tomviz._internal import convert_to_vtk_data_object
        data = convert_to_vtk_data_object(value)
        self._operator._operator_wrapper.progress_data = data

    # Write-only property to update child data
    data = property(fset=_data)


class Operator(object):
    """
    The base operator class from which all operators should be derived.
    """
    def __new__(cls, *args, **kwargs):
        obj = super(Operator, cls).__new__(cls)
        obj.progress = Progress(obj)

        return obj

    def transform_scalars(self, data):
        """
        This method should be overridden by subclasses to implement the
        operations the operator should perform.
        """
        raise NotImplementedError('Must be implemented by subclass')

    def transform(self, data):
        """
        This method should be overridden by subclasses to implement the
        operations the operator should perform.
        """
        raise NotImplementedError('Must be implemented by subclass')


class CancelableOperator(Operator):
    """
    A cancelable operator allows the user to interrupt the execution of the
    operator. The canceled property can be using in the transform(...)
    method to break out when the operator is canceled. The basic structure of
    the transform(...) might look something like this:

    def transform(self, data):
        while(not self.canceled):
            # Do work

    """
    @property
    def canceled(self):
        """
        :returns True if the operator has been canceled, False otherwise.
        """
        return self._operator_wrapper.canceled


class CompletableOperator(CancelableOperator):
    """
    A completable operator allows a user to interrupt the execution of
    an operator using either "cancel" or "complete". The
    completable property can be used in the transform(...) method to break out
    when an operator is finished early, like if an iterative algorithm is a
    reasonable quality before the designated iterations are reached. Use
    similar to "cancel", but be sure to return data.
    """
    @property
    def completed(self):
        """
        :returns True if the operator is early completed (from Button),
        False otherwise
        """
        return self._operator_wrapper.completed



def transform(dataset, sigma=2.0):
    """Apply a Gaussian filter to volume dataset."""
    """Gaussian Filter blurs the image and reduces the noise and details."""

    import scipy.ndimage
    import numpy as np

    array = dataset.active_scalars

    # Transform the dataset.
    result = np.empty_like(array)
    scipy.ndimage.filters.gaussian_filter(array, sigma, output=result)

    # Set the result as the new scalars.
    dataset.active_scalars = result



# -*- coding: utf-8 -*-

###############################################################################
# This source file is part of the Tomviz project, https://tomviz.org/.
# It is released under the 3-Clause BSD License, see "LICENSE".
###############################################################################
import inspect
import sys
import os
import fnmatch
import imp
import json
import traceback

import tomviz
import tomviz.operators


def in_application():
    return os.environ.get('TOMVIZ_APPLICATION', False)


if in_application():
    import tomviz._wrapping
    from vtk import vtkDataObject


def require_internal_mode():
    if not in_application():
        func_name = str(inspect.currentframe().f_back.f_code.co_name)
        raise Exception('Cannot call ' + func_name + ' in external mode')


def delete_module(name):
    if name in sys.modules:
        del sys.modules[name]


def find_operator_class(transform_module):
    operator_class = None
    classes = inspect.getmembers(transform_module, inspect.isclass)
    for (name, cls) in classes:
        if issubclass(cls, tomviz.operators.Operator):
            if operator_class is not None:
                raise Exception('Multiple operators define in module, only '
                                'one operator can be defined per module.')

            operator_class = cls

    return operator_class


def _find_function(module, function_name):
    # Finds a function in the module with a given "function_name"
    # Returns `None` if it is not found
    functions = inspect.getmembers(module, inspect.isfunction)
    for (name, func) in functions:
        if name == function_name:
            return func


def find_transform_from_module(transform_module):
    # This tries to first find transform(), and then transform_scalars()
    f = _find_function(transform_module, 'transform')
    if f is None:
        f = _find_function(transform_module, 'transform_scalars')

    return f


def is_cancelable(transform_module):
    cls = find_operator_class(transform_module)

    if cls is None:
        function = find_transform_from_module(transform_module)

    if cls is None and function is None:
        raise Exception('Unable to locate function or operator class.')

    return cls is not None and issubclass(cls,
                                          tomviz.operators.CancelableOperator)


def is_completable(transform_module):
    cls = find_operator_class(transform_module)

    if cls is None:
        function = find_transform_from_module(transform_module)
        if not function:
            raise Exception('Unable to locate function or operator class.')

    return cls is not None and issubclass(
        cls,
        tomviz.operators.CompletableOperator
    )


def find_transform_function(transform_module, op=None):

    transform_function = find_transform_from_module(transform_module)
    if transform_function is None:
        cls = find_operator_class(transform_module)
        if cls is None:
            raise Exception('Unable to locate transform function.')

        # We call __new__ and __init__ manually here so we can inject the
        # wrapper OperatorPython instance before __init__ is called so that
        # any code in __init__ can access the wrapper.
        o = cls.__new__(cls)
        if op is not None:
            # Set the wrapped OperatorPython instance
            o._operator_wrapper = tomviz._wrapping.OperatorPythonWrapper(op)
        cls.__init__(o)

        transform_function = None
        if _operator_method_was_implemented(o, 'transform'):
            transform_function = o.transform
        elif _operator_method_was_implemented(o, 'transform_scalars'):
            transform_function = o.transform_scalars

    if transform_function is None:
        raise Exception('Unable to locate transform function.')

    return transform_function


def _load_module(operator_dir, python_file):
    module_name, _ = os.path.splitext(python_file)
    fp, pathname, description = imp.find_module(module_name, [operator_dir])
    module = imp.load_module(module_name, fp, pathname, description)

    return module


def _has_operator(module):
    return find_transform_from_module(module) is not None or \
        find_operator_class(module) is not None


def _operator_description(operator_dir, filename):
    name, _ = os.path.splitext(filename)
    description = {
        'label': name,
        'pythonPath': os.path.join(operator_dir, filename),
    }

    has_operator = False
    # Load the module and see if they are valid operators
    try:
        module = _load_module(operator_dir, filename)
        has_operator = _has_operator(module)
    except Exception:
        description['loadError'] = traceback.format_exc()

    description['valid'] = has_operator

    # See if we have a JSON file
    json_filepath = os.path.join(operator_dir, '%s.json' % name)
    if os.path.exists(json_filepath):
        description['jsonPath'] = json_filepath
        # Extract the label from the JSON
        try:
            with open(json_filepath) as fp:
                operator_json = json.load(fp)
                description['label'] = operator_json['label']
        except Exception:
            description['loadError'] = traceback.format_exc()
            description['valid'] = False

    return description


def find_operators(operator_dir):
    # First look for the python files
    python_files = fnmatch.filter(os.listdir(operator_dir), '*.py')
    operator_descriptions = []
    for python_file in python_files:
        operator_descriptions.append(
            _operator_description(operator_dir, python_file)
        )

    return operator_descriptions


def _operator_method_was_implemented(obj, method):
    # It would be nice if there were an easier way to do this, but
    # I am not currently aware of an easier way.
    bases = list(inspect.getmro(type(obj)))
    # We know operator has this attribute, remove it
    bases.remove(tomviz.operators.Operator)

    for base in bases:
        if method in vars(base):
            return True

    return False


def convert_to_dataset(data):
    # This method will extract/convert certain data types to a dataset

    if in_application():
        from tomviz.internal_dataset import Dataset
    else:
        from tomviz.external_dataset import Dataset

    if isinstance(data, Dataset):
        # It is already a dataset
        return data

    if in_application():
        if isinstance(data, vtkDataObject):
            # Make a new dataset object with no Datasource
            return Dataset(data, None)

    msg = 'Cannot convert type to Dataset: ' + str(type(data))
    raise Exception(msg)


def convert_to_vtk_data_object(data):
    # This method will extract/convert certain data types to a vtkDataObject

    from tomviz.internal_dataset import Dataset

    if isinstance(data, vtkDataObject):
        # It is already a vtkDataObject
        return data

    if isinstance(data, Dataset):
        # Should be stored in _data_object
        return data._data_object

    msg = 'Cannot convert type to vtkDataObject: ' + str(type(data))
    raise Exception(msg)


def with_vtk_dataobject(f):
    # A decorator to automatically convert the first argument to
    # a vtkDataObject. This also confirms we are running internally.

    def wrapped(*args, **kwargs):
        if not in_application():
            name = f.__name__
            raise Exception('Cannot call ' + name + ' in external mode')

        dataobject = convert_to_vtk_data_object(args[0])
        args = (dataobject, *args[1:])
        return f(*args, **kwargs)

    return wrapped


def with_dataset(f):
    # A decorator to automatically convert the first argument to
    # a Dataset.

    def wrapped(*args, **kwargs):
        dataset = convert_to_dataset(args[0])
        args = (dataset, *args[1:])
        return f(*args, **kwargs)

    return wrapped



import tomviz._realtime.ctvlib as ctvlib
from tomviz._realtime import logger, wbp
from tomviz._realtime import pytvlib
import tomviz.operators
import time


class RealTimeTomography(tomviz.operators.CompletableOperator):

    def transform(self, dataset, localDirectory=None, alg=None,
                  maxIter=None, fileExt=None):

        # Parse Inputs
        fileExtensions = ('dm4', 'dm3', 'ser')
        algorithms = ('ART', 'randART', 'SIRT', 'WBP')
        alg = algorithms[alg]

        # Boolean to Determine Whether Algorithm is ART.
        artBool = alg == 'ART' or alg == 'randART'

        # Create child for recon
        child = dataset.create_child_dataset()

        # Logger to Read Directory
        tomoLogger = logger.Logger(localDirectory, fileExtensions[fileExt])

        # Keep Checking Directory until a projection is read.
        while True:
            if tomoLogger.monitor():
                break

        (Nslice, Nray, Nproj) = tomoLogger.logTiltSeries.shape

        # Initialize C++ Object..
        if alg != 'WBP':
            tomo = ctvlib.ctvlib(Nslice, Nray, Nproj)
        else:
            tomo = wbp.WBP(Nslice, Nray, Nproj)
            beta = 0
            maxIter = Nslice
        tomoLogger.load_tilt_series(tomo, alg)

        # Progress Bar
        self.progress.maximum = maxIter

        # Generate measurement matrix (if Iterative Algorithm)
        self.progress.message = 'Generating measurement matrix'
        pytvlib.initialize_algorithm(tomo, alg, Nray, tomoLogger.logTiltAngles)

        # Descent Parameter Initialization
        if alg == 'SIRT':
            beta = 1/tomo.lipschits()
        elif artBool:
            beta0 = 0.5
            beta = beta0
            betaRed = 0.99

        # Dynamic Tilt Series Loop.
        runExperiment = True
        while runExperiment:

            t0 = time.time()
            counter = 1
            etcMessage = 'Estimated time to complete: n/a'

            #Main Reconstruction Loop
            for jj in range(maxIter):

                # Check to see experiment is complete
                if self.completed or self.canceled:
                    runExperiment = False
                    break

                self.progress.message = 'Reconstructing Tilt Angles (%d'\
                                        ' Sampled): Iteration No.%d/%d. '\
                    % (tomoLogger.logTiltAngles.shape[0], jj + 1,
                        maxIter) + etcMessage

                # Run Reconstruction Algorithm
                pytvlib.run(tomo, alg, beta, jj)

                # Decay Descent Parameter if ART.
                if artBool:
                    beta *= betaRed

                # Progress Bar
                self.progress.value = jj + 1
                (etcMessage, counter) = pytvlib.timer(t0, counter, maxIter)

            # Return a Current Iterate
            self.progress.message = 'Updating Tomogram Visualization'
            if alg != 'WBP':
                child.active_scalars = pytvlib.get_recon((Nslice, Nray), tomo)
            else:
                child.active_scalars = tomo.recon
            self.progress.data = child

            # Run Logger to see how many projections were collected since
            # last check.
            if tomoLogger.monitor():

                # Update tomo (C++) with new projections / tilt Angles.
                self.progress.message = 'Generating measurement matrix'
                pytvlib.initialize_algorithm(tomo, alg, Nray,
                                             tomoLogger.logTiltAngles, 1)
                tomoLogger.load_tilt_series(tomo, alg)

                # Recalculate Lipschitz Constant or Reset Descent Parameter
                if alg == 'SIRT':
                    beta = 1/tomo.lipschits()
                elif artBool:
                    beta = beta0

        # One last update of the child data.
        child.active_scalars = pytvlib.get_recon((Nslice, Nray), tomo)
        self.progress.data = child

        returnValues = {}
        returnValues["reconstruction"] = child
        return returnValues



import tomviz.operators
import tomviz.utils

import os
from enum import Enum
import numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import dijkstra


class DistanceMethod(Enum):
    Eucledian = 0
    CityBlock = 1
    ChessBoard = 2


class PropagationDirection(Enum):
    Xpos = 0
    Xneg = 1
    Ypos = 2
    Yneg = 3
    Zpos = 4
    Zneg = 5


class OperatorStages(Enum):
    GraphGeneration = 0
    GraphTraversal = 1


def coord_iterator(extent):
    if len(extent) == 1:
        for i in range(extent[0][0], extent[0][1]):
            yield (i, )
    elif len(extent) > 1:
        for i in range(extent[0][0], extent[0][1]):
            for c in coord_iterator(extent[1:]):
                yield (i, ) + c


def neighbor_iterator(ndim):
    for c in coord_iterator(((-1, 2),) * ndim):
        if any(c):
            yield c


def get_distance_function(method):
    if method == DistanceMethod.Eucledian:
        import math

        def distance_fn(vec):
            return math.sqrt(sum(x * x for x in vec))

        return distance_fn

    elif method == DistanceMethod.ChessBoard:

        def distance_fn(vec):
            return 1

        return distance_fn

    elif method == DistanceMethod.CityBlock:

        def distance_fn(vec):
            if sum(abs(x) for x in vec) > 1:
                return None
            else:
                return 1

        return distance_fn

    else:
        raise Exception("Unknown distance method %s" % method)


def volume_to_graph(volume, phase, method, update_progress=None):
    invalid_node_idx = -1
    # The distance function
    distance = get_distance_function(method)
    # Map voxel coordinate to node index
    node_map = {}
    # Map node index to voxel coordinate
    inv_node_map = {}
    # Map voxel coordinate to node index (array form).
    # This is useful to easily slice when doing analysis.
    # We can do without if memory becomes an issue.
    node_map_array = np.empty(shape=volume.shape, dtype=np.int32)
    node_map_array.fill(invalid_node_idx)
    # Weighted (i.e. distance) connections between nodes
    edges = {}
    # Edges between auxiliary face nodes and actual nodes on the first slice
    # (x+, x-, y+, y-, z+, z-)
    aux_edges = {}

    # Reserve node_idx 0 to ndim * 2 - 1 for the auxiliary nodes.
    # These are nodes at the beginning/end of the x/y direction
    # In the propagation this are the distances we are calculating from
    node_idx = volume.ndim * 2

    # Map node indices to positions in the volume
    extent = tuple((0, s) for s in volume.shape)
    for coord in coord_iterator(extent):
        if volume[coord] == phase:
            node_map_array[coord] = node_idx
            inv_node_map[node_idx] = coord
            node_map[coord] = node_idx
            node_idx += 1

    # Find edges between the nodes

    delta_coordinates = [d for d in neighbor_iterator(volume.ndim)]
    delta_distances = [distance(c) for c in delta_coordinates]

    n_nodes = len(node_map)

    # Update every 5%
    update_every = max(n_nodes // 20, 1)

    for i, (node_idx, node_coord) in enumerate(inv_node_map.items()):
        if i % update_every == 0 and update_progress is not None:
            update_progress(i / n_nodes)

        for delta_coord, delta_dist in zip(delta_coordinates, delta_distances):
            neighbor_coord = tuple(c + d
                                   for c, d in zip(node_coord, delta_coord))
            neighbor_idx = node_map.get(neighbor_coord)

            if neighbor_idx is None:
                continue

            if delta_dist is not None:
                # Reduce memory footprint by only saving edges once
                edge_key = (min(node_idx, neighbor_idx),
                            max(node_idx, neighbor_idx))
                edges[edge_key] = delta_dist

    # Add edges between aux nodes at the faces of the volume
    for i in range(volume.ndim):
        for j in range(2):
            node_idx = i * 2 + j
            face_slice = [slice(None)] * volume.ndim
            face_slice[i] = -j
            face_slice = tuple(face_slice)
            for neighbor_idx in node_map_array[face_slice].flatten():
                if neighbor_idx != invalid_node_idx:
                    aux_edges[(node_idx, neighbor_idx)] = 1

    if update_progress is not None:
        update_progress(1)

    return node_map, inv_node_map, node_map_array, edges, aux_edges


def edges_to_sparse_matrix(edges, aux_edges, aux_node_idx, n_nodes, ndim):
    n_edges = len(edges)
    n_aux_edges = 0

    for (from_idx, _), _ in aux_edges.items():
        if from_idx == aux_node_idx:
            n_aux_edges += 1

    row = np.empty(shape=(n_edges + n_aux_edges,), dtype=np.int32)
    col = np.empty(shape=(n_edges + n_aux_edges,), dtype=np.int32)
    data = np.empty(shape=(n_edges + n_aux_edges,), dtype=np.float32)

    i = 0
    for (from_idx, to_idx), distance in edges.items():
        row[i] = from_idx
        col[i] = to_idx
        data[i] = distance
        i += 1

    for (from_idx, to_idx), distance in aux_edges.items():
        if from_idx == aux_node_idx:
            row[i] = from_idx
            col[i] = to_idx
            data[i] = distance
            i += 1

    assert(i == n_edges + n_aux_edges)

    total_nodes = n_nodes + 2 * ndim
    sparse_edge_matrix = csr_matrix((data, (row, col)),
                                    shape=(total_nodes, total_nodes))

    return sparse_edge_matrix


def distance_matrix_to_volume(inv_node_map, dist_matrix, shape):
    volume = np.empty(shape=shape, dtype=np.float32)
    unreachable_scalar_value = -1
    volume.fill(unreachable_scalar_value)
    for node_idx, node_coord in inv_node_map.items():
        volume[node_coord] = dist_matrix[node_idx]
    volume[volume == np.inf] = unreachable_scalar_value
    volume[volume == np.nan] = unreachable_scalar_value
    return volume


def get_slice_scalars(volume, propagation_direction, slice_number):
    axis_idx = propagation_direction // 2
    start = 0
    increment = 1
    if propagation_direction % 2 == 1:
        increment = -1
        start = volume.shape[axis_idx] - 1

    s = start + increment * slice_number
    assert(s >= 0 and s < volume.shape[axis_idx])

    slice_obj = [slice(None)] * volume.ndim
    slice_obj[axis_idx] = s
    slice_obj = tuple(slice_obj)
    scalars = volume[slice_obj]

    return scalars


def calculate_avg_path_length(volume, propagation_direction):
    unreachable_scalar_value = -1
    axis_idx = propagation_direction // 2
    n_slices = volume.shape[axis_idx]
    column_names = ["Linear Distance", "Actual Distance"]
    table_data = np.empty(shape=(n_slices, 2))
    for i in range(n_slices):
        scalars = get_slice_scalars(volume, propagation_direction, i)
        scalars = np.extract(scalars != unreachable_scalar_value, scalars)
        table_data[i, 0] = i + 1
        table_data[i, 1] = scalars.mean()

    return column_names, table_data


def calculate_tortuosity(straight_distance, path_length):
    column_names = ["Scale", "End", "Average", "Slope"]
    table_data = np.empty(shape=(1, 4))

    # Scale Tortuosity
    m, _ = np.polyfit(straight_distance, path_length, deg=1)
    table_data[0, 0] = m

    # End Tortuosity
    table_data[0, 1] = path_length[-1] / straight_distance[-1]

    # Average Tortuosity
    table_data[0, 2] = np.mean(path_length / straight_distance)

    # Slope Tortuosity
    # TODO: implement Slope Tortuosity
    table_data[0, 3] = -1

    return column_names, table_data


def calculate_tortuosity_distribution(volume, propagation_direction):
    unreachable_scalar_value = -1
    axis_idx = propagation_direction // 2
    n_slices = volume.shape[axis_idx]
    last_slice = n_slices - 1

    scalars = get_slice_scalars(volume, propagation_direction, last_slice)
    scalars = np.extract(scalars != unreachable_scalar_value, scalars)
    linear_distance = last_slice + 1
    scalars /= linear_distance

    n_bins = 100
    column_names = ["Tortuosity", "Occurrence"]
    table_data = np.empty(shape=(n_bins, 2))
    bins = np.linspace(1, 6, num=n_bins + 1)
    occurrence, _ = np.histogram(scalars, bins)
    for i in range(n_bins):
        table_data[i, 0] = bins[i]
        table_data[i, 1] = occurrence[i]

    return column_names, table_data


def get_update_progress_fn(progress, stage):
    GRAPH_GENERATION_FRACTION = 0.9
    OTHER_FRACTION = 1 - GRAPH_GENERATION_FRACTION

    if stage == OperatorStages.GraphGeneration:
        def update_progress(value):
            progress.value = round(progress.maximum * value *
                                   GRAPH_GENERATION_FRACTION)

        return update_progress

    elif stage == OperatorStages.GraphTraversal:
        def update_progress(value):
            progress.value = round(progress.maximum *
                                   (value * OTHER_FRACTION +
                                    GRAPH_GENERATION_FRACTION))

        return update_progress

    else:
        def update_progress(value):
            progress.value = round(progress.maximum * value)

        return update_progress


class TortuosityOperator(tomviz.operators.CancelableOperator):
    """Distance propagation method for calculating tortuosity.
    https://doi.org/10.1016/j.jpowsour.2013.10.026
    """
    def transform(self, dataset, phase=1,
                  distance_method=DistanceMethod.Eucledian,
                  propagation_direction=PropagationDirection.Xpos,
                  save_to_file=False, output_folder=""):
        """Operator transform method

        Args:
            phase (int): the scalar value in the dataset that is considered
                         a pore
            distance_method (enum): the distance method to calculate distance
                                    between nodes
            propagation_direction (enum): the face from which distances are
                                          calculated (X+, X-, Y+, etc.)
            save_to_file (bool): save the detailed output of the operator to
                                 files. If set to True, propagate along all
                                 six directions and save the results, but only
                                 display results for one in the application.
            output_folder (str): the path to the folder where the optional
                                 output files are written to
        """
        distance_method = DistanceMethod(distance_method)
        propagation_direction = PropagationDirection(propagation_direction)
        scalars = dataset.active_scalars

        if scalars is None:
            raise RuntimeError("No scalars found!")

        if save_to_file and not os.access(output_folder, os.W_OK):
            import warnings
            save_to_file = False
            warnings.warn(
                "Unable to write to destination folder %s" % output_folder)

        if save_to_file:
            propagation_directions = list(PropagationDirection)
        else:
            propagation_directions = [propagation_direction]

        self.progress.maximum = 100

        graph_generation_update_progress_fn = get_update_progress_fn(
            self.progress, OperatorStages.GraphGeneration)
        graph_traversal_update_progress_fn = get_update_progress_fn(
            self.progress, OperatorStages.GraphTraversal)

        self.progress.message = "Converting volume to graph..."
        node_map, inv_node_map, node_map_array, edges, aux_edges = (
            volume_to_graph(
                scalars, phase, distance_method,
                graph_generation_update_progress_fn)
        )

        n_directions = len(propagation_directions)

        return_values = {}

        for i, direction in enumerate(propagation_directions):
            self.progress.message = "Propagating along %s" % direction.name
            graph_traversal_update_progress_fn(i / n_directions)

            csgraph = edges_to_sparse_matrix(
                edges, aux_edges, direction.value,
                len(inv_node_map), scalars.ndim)

            graph_traversal_update_progress_fn(i / n_directions + 0.33)

            dist_matrix = dijkstra(csgraph, directed=False,
                                   indices=direction.value)

            graph_traversal_update_progress_fn(i / n_directions + 0.9)

            # Generate the distance map

            result = distance_matrix_to_volume(inv_node_map, dist_matrix,
                                               scalars.shape)

            if save_to_file:
                filename = "distance_map_%s.npy" % direction.name
                np.save(os.path.join(output_folder, filename), result)

            if direction == propagation_direction:
                dataset.active_scalars = result

            # Calculate the average path length per slice

            column_names, table_data = calculate_avg_path_length(
                result, direction.value)

            if save_to_file:
                filename = "path_length_%s.csv" % direction.name
                np.savetxt(os.path.join(output_folder, filename), table_data,
                           delimiter=", ", header=", ".join(column_names))

            if direction == propagation_direction:
                table = tomviz.utils.make_spreadsheet(column_names, table_data)
                return_values["path_length"] = table

            # Calculate the tortuosity (4 different ways)

            column_names, table_data = calculate_tortuosity(
                table_data[:, 0], table_data[:, 1])

            if save_to_file:
                filename = "tortuosity_%s.csv" % direction.name
                np.savetxt(os.path.join(output_folder, filename), table_data,
                           delimiter=", ", header=", ".join(column_names))

            if direction == propagation_direction:
                table = tomviz.utils.make_spreadsheet(column_names, table_data)
                return_values["tortuosity"] = table

            # Calculate the tortuosity distribution of the last slice

            column_names, table_data = calculate_tortuosity_distribution(
                result, direction.value)

            if save_to_file:
                filename = "tortuosity_distribution_%s.csv" % direction.name
                np.savetxt(os.path.join(output_folder, filename), table_data,
                           delimiter=", ", header=", ".join(column_names))

            if direction == propagation_direction:
                table = tomviz.utils.make_spreadsheet(column_names, table_data)
                return_values["tortuosity_distribution"] = table

        return return_values



import tomviz.operators


def median_filter(operator, step_pct, input_image):
    import itk
    from tomviz import itkutils

    median_filter = \
        itk.MedianImageFilter.New(Input=input_image)
    dimension = input_image.GetImageDimension()
    median_radius = itk.Size[dimension]()
    median_radius.Fill(1)
    median_filter.SetRadius(median_radius)

    operator.progress.message = "Median filter"
    progress = operator.progress.value
    itkutils.observe_filter_progress(operator, median_filter,
                                     progress, progress + next(step_pct))
    try:
        median_filter.Update()
    except RuntimeError:
        return

    return median_filter.GetOutput()


def unsharp_mask(operator, step_pct, input_image):
    import itk
    from tomviz import itkutils
    import numpy as np

    enhancer = \
        itk.UnsharpMaskImageFilter.New(Input=input_image)
    enhancer.SetAmount(3.0)
    spacing = input_image.GetSpacing()
    enhancer.SetSigma(2.0 * np.max(spacing))

    operator.progress.message = "Unsharp mask"
    progress = operator.progress.value
    itkutils.observe_filter_progress(operator, enhancer,
                                     progress, progress + next(step_pct))
    try:
        enhancer.Update()
    except RuntimeError:
        return

    return enhancer.GetOutput()


def get_distance(operator, step_pct, input_image):
    import itk
    from tomviz import itkutils

    distance_mapper = \
        itk.SignedMaurerDistanceMapImageFilter.New(Input=input_image)
    distance_mapper.SetInsideIsPositive(True)
    distance_mapper.SetSquaredDistance(False)
    distance_mapper.SetUseImageSpacing(True)

    operator.progress.message = "Distance map"
    progress = operator.progress.value
    itkutils.observe_filter_progress(operator, distance_mapper,
                                     progress, progress + next(step_pct))
    try:
        distance_mapper.Update()
    except RuntimeError:
        return

    return distance_mapper.GetOutput()


def invert(operator, step_pct, input_image):
    import itk
    from tomviz import itkutils

    inverter = \
        itk.NotImageFilter.New(Input=input_image)

    operator.progress.message = "Invert  threshold"
    progress = operator.progress.value
    itkutils.observe_filter_progress(operator, inverter,
                                     progress, progress + next(step_pct))
    try:
        inverter.Update()
    except RuntimeError:
        return

    return inverter.GetOutput()


def threshold(operator, step_pct, input_image):
    import itk
    from tomviz import itkutils
    import itkExtras
    import itkTypes

    otsu_filter = \
        itk.OtsuMultipleThresholdsImageFilter.New(Input=input_image)
    otsu_filter.SetNumberOfThresholds(1)
    operator.progress.message = "Otsu threshold"
    progress = operator.progress.value
    itkutils.observe_filter_progress(operator, otsu_filter,
                                     progress, progress + next(step_pct))

    try:
        otsu_filter.Update()
    except RuntimeError:
        return

    # Cast threshold output to an integral type if needed.
    input_image_type = type(input_image)
    voxel_type = itkExtras.template(input_image_type)[1][0]
    if voxel_type is itkTypes.F or voxel_type is itkTypes.D:
        operator.progress.message = "Casting output to integral type"

        # Unsigned char supports 256 labels, or 255 threshold levels.
        # This should be sufficient for all but the most unusual use
        # cases.
        dimension = input_image_type.GetImageDimension()
        py_buffer_type = itk.Image[itk.UC, dimension]
        caster = itk.CastImageFilter[input_image_type,
                                     py_buffer_type].New()
        caster.SetInput(thresholded)
        progress = operator.progress.value
        itkutils.observe_filter_progress(operator, caster,
                                         progress, progress + next(step_pct))

        try:
            caster.Update()
        except RuntimeError:
            return

        thresholded = caster.GetOutput()

    thresholded = otsu_filter.GetOutput()
    return thresholded


def watershed(operator, step_pct, input_image, level):
    import itk
    from tomviz import itkutils

    dimension = input_image.GetImageDimension()
    watershed_input_type = itk.Image[itk.F, dimension]
    watershed_output_type = itk.Image[itk.US, dimension]
    caster = itk.CastImageFilter[type(input_image), watershed_input_type].New()
    caster.SetInput(input_image)

    watershed_filter = \
        itk.MorphologicalWatershedImageFilter[
            watershed_input_type,
            watershed_output_type].New(Input=caster.GetOutput())
    watershed_filter.SetLevel(level)
    watershed_filter.SetMarkWatershedLine(True)
    watershed_filter.SetFullyConnected(True)

    operator.progress.message = "Watershed filter"
    progress = operator.progress.value
    itkutils.observe_filter_progress(operator, watershed_filter,
                                     progress, progress + next(step_pct))
    try:
        watershed_filter.Update()
    except RuntimeError:
        return

    return watershed_filter.GetOutput()


def apply_mask(operator, step_pct, input_image1, input_image2):
    import itk

    dimension = input_image1.GetImageDimension()
    output_type = itk.Image[itk.US, dimension]
    caster1 = itk.CastImageFilter[type(input_image1), output_type].New()
    caster1.SetInput(input_image1)
    caster2 = itk.CastImageFilter[type(input_image2), output_type].New()
    caster2.SetInput(input_image2)
    mask_filter = \
        itk.MultiplyImageFilter.New(Input1=caster1.GetOutput(),
                                    Input2=caster2.GetOutput())

    operator.progress.message = "Apply mask"
    operator.progress.value += next(step_pct)
    try:
        mask_filter.Update()
    except RuntimeError:
        return

    return mask_filter.GetOutput()


def morphological_closing(operator, step_pct, input_image, structuring_element):
    import itk
    from tomviz import itkutils

    closer = \
        itk.GrayscaleMorphologicalClosingImageFilter.New(Input=input_image)
    closer.SetKernel(structuring_element)
    operator.progress.message = "Morphological closing"
    progress = operator.progress.value
    itkutils.observe_filter_progress(operator, closer,
                                     progress, progress + next(step_pct))

    try:
        closer.Update()
    except RuntimeError:
        return

    return closer.GetOutput()


def encapsulate(operator, step_pct, input_image, mask, structuring_element):
    import itk
    from tomviz import itkutils

    dilater = \
        itk.GrayscaleDilateImageFilter.New(Input=input_image)
    dilater.SetKernel(structuring_element)
    operator.progress.message = "Encapsulate"
    progress = operator.progress.value
    nextprogress = progress + next(step_pct)
    itkutils.observe_filter_progress(operator, dilater,
                                     progress, nextprogress)

    xor_filter = \
        itk.XorImageFilter.New(Input1=mask, Input2=dilater.GetOutput())
    progress = nextprogress
    nextprogress = progress + next(step_pct)
    itkutils.observe_filter_progress(operator, xor_filter,
                                     progress, nextprogress)

    or_filter = \
        itk.OrImageFilter.New(Input1=input_image, Input2=xor_filter.GetOutput())
    progress = nextprogress
    nextprogress = progress + next(step_pct)
    itkutils.observe_filter_progress(operator, or_filter,
                                     progress, nextprogress)

    try:
        or_filter.Update()
    except RuntimeError:
        return

    return or_filter.GetOutput()


def opening_by_reconstruction(operator, step_pct, input_image,
                              structuring_element):
    import itk
    from tomviz import itkutils

    opening = \
        itk.OpeningByReconstructionImageFilter.New(Input=input_image)
    opening.SetKernel(structuring_element)

    operator.progress.message = "Opening by reconstruction"
    progress = operator.progress.value
    itkutils.observe_filter_progress(operator, opening,
                                     progress, progress + next(step_pct))

    try:
        opening.Update()
    except RuntimeError:
        return

    return opening.GetOutput()


class SegmentPores(tomviz.operators.CancelableOperator):

    def transform(self, dataset, minimum_radius=0.5, maximum_radius=6.):
        """Segment pores. The pore size must be greater than the minimum radius
        and less than the maximum radius.  Pores will be separated according to
        the minimum radius."""

        # Initial progress
        self.progress.value = 0
        self.progress.maximum = 100

        # Approximate percentage of work completed after each step in the
        # transform
        step_pct = iter([5, 5, 5, 5, 5, 5, 5, 5, 5, 30, 10, 5, 5, 5])

        try:
            import itk
            from tomviz import itkutils
            import numpy as np
        except Exception as exc:
            print("Could not import necessary module(s)")
            raise exc

        # Return values
        returnValues = None

        # Add a try/except around the ITK portion. ITK exceptions are
        # passed up to the Python layer, so we can at least report what
        # went wrong with the script, e.g,, unsupported image type.
        try:
            self.progress.message = "Converting data to ITK image"

            # Get the ITK image
            itk_input_image = itkutils.dataset_to_itk_image(dataset)
            self.progress.value = next(step_pct)

            # Reduce noise
            smoothed = median_filter(self, step_pct, itk_input_image)

            # Enhance pore contrast
            enhanced = unsharp_mask(self, step_pct, smoothed)

            thresholded = threshold(self, step_pct, enhanced)

            dimension = itk_input_image.GetImageDimension()
            spacing = itk_input_image.GetSpacing()
            closing_radius = itk.Size[dimension]()
            closing_radius.Fill(1)
            for dim in range(dimension):
                radius = int(np.round(maximum_radius / spacing[dim]))
                if radius > closing_radius[dim]:
                    closing_radius[dim] = radius
            StructuringElementType = itk.FlatStructuringElement[dimension]
            structuring_element = \
                StructuringElementType.Ball(closing_radius)
            particle_mask = morphological_closing(self, step_pct, thresholded,
                                                  structuring_element)

            encapsulated = encapsulate(self, step_pct, thresholded,
                                       particle_mask, structuring_element)

            distance = get_distance(self, step_pct, encapsulated)

            segmented = watershed(self, step_pct, distance, minimum_radius)

            inverted = invert(self, step_pct, thresholded)

            segmented.DisconnectPipeline()
            inverted.DisconnectPipeline()
            separated = apply_mask(self, step_pct, segmented, inverted)

            separated.DisconnectPipeline()
            particle_mask.DisconnectPipeline()
            in_particles = apply_mask(self, step_pct, separated, particle_mask)

            opening_radius = itk.Size[dimension]()
            opening_radius.Fill(1)
            for dim in range(dimension):
                radius = int(np.round(minimum_radius / spacing[dim]))
                if radius > opening_radius[dim]:
                    opening_radius[dim] = radius
            structuring_element = \
                StructuringElementType.Ball(opening_radius)
            opened = opening_by_reconstruction(self, step_pct, in_particles,
                                               structuring_element)

            self.progress.message = "Saving results"

            label_map_dataset = dataset.create_child_dataset()
            itkutils.set_itk_image_on_dataset(opened, label_map_dataset)

            # Set up dictionary to return operator results
            returnValues = {}
            returnValues["label_map"] = label_map_dataset

        except Exception as exc:
            print("Problem encountered while running %s" %
                  self.__class__.__name__)
            raise exc

        return returnValues



def transform(dataset):
    """Downsample tilt images by a factor of 2"""

    from tomviz import utils
    import scipy.ndimage
    import numpy as np
    import warnings

    array = dataset.active_scalars

    zoom = (0.5, 0.5, 1)
    result_shape = utils.zoom_shape(array, zoom)
    result = np.empty(result_shape, array.dtype, order='F')
    # Downsample the dataset x2 using order 1 spline (linear)
    warnings.filterwarnings('ignore', '.*output shape of zoom.*')
    scipy.ndimage.interpolation.zoom(array, zoom,
                                     output=result,
                                     order=1,
                                     mode='constant',
                                     cval=0.0, prefilter=False)

    # Set the result as the new scalars.
    dataset.active_scalars = result



def transform(dataset, label_value=1):
    """Computes the principal axes of an object with the label value passed in
    as the parameter label_value. The principal axes are added to the field
    data of the dataset as an array of 3-component tuples named 'PrincipalAxes'.
    The principal axis tuples are stored in order of length from longest to
    shortest. The center of the object is also added to the field data as an
    array named 'Center' with a single 3-component tuple.

    This operator uses principal components analysis of the labeled point
    locations to determine the principal axes.
    """

    from tomviz import utils

    (axes, center) = utils.label_object_principal_axes(dataset, label_value)

    print('Axes is:', axes)
    print('Center is:', center)

    # These set the field data on the vtkDataObject
    utils.set_principal_axes(dataset, axes)
    utils.set_center(dataset, center)



def transform(dataset, threshold=None):
    """Remove bad pixels in tilt series."""

    import scipy.ndimage
    import numpy as np

    tiltSeries = dataset.active_scalars.astype(np.float32)

    for i in range(tiltSeries.shape[2]):
        I = tiltSeries[:, :, i]
        I_pad = np.lib.pad(I, (1, 1), 'edge')

        # calculate standard deviation in a 3 x 3 window
        averageI2 = scipy.ndimage.filters.uniform_filter(I_pad ** 2)
        averageI = scipy.ndimage.filters.uniform_filter(I_pad)
        std = np.sqrt(abs(averageI2 - averageI**2))[1:-1, 1:-1]

        medianI = scipy.ndimage.filters.median_filter(I_pad, 2)[1:-1, 1:-1]

        #identify bad pixels
        badPixelsMask = abs(I - medianI) > std * threshold

        I[badPixelsMask] = medianI[badPixelsMask]
        tiltSeries[:, :, i] = I

    # Set the result as the new scalars.
    dataset.active_scalars = tiltSeries



import numpy as np
import tomviz.operators


class RandomTiltSeriesShiftOperator(tomviz.operators.CancelableOperator):

    def transform(self, dataset, maxShift=1):
        """Apply random integer shifts to tilt images"""

        tiltSeries = dataset.active_scalars
        if tiltSeries is None:
            raise RuntimeError("No scalars found!")

        self.progress.maximum = tiltSeries.shape[2]
        step = 0
        for i in range(tiltSeries.shape[2]):
            if self.canceled:
                return
            shifts = (np.random.rand(2) * 2 - 1) * maxShift
            tiltSeries[:, :, i] = np.roll(
                tiltSeries[:, :, i], int(shifts[0]), axis=0)
            tiltSeries[:, :, i] = np.roll(
                tiltSeries[:, :, i], int(shifts[1]), axis=1)
            step += 1
            self.progress.value = step

        dataset.active_scalars = tiltSeries



from . import channel

class layer:
    """Layer Class
    A layer is a collection of values that comprise an element of an image. 
    A layer is an integer size (wxh) that is equal to or lesser than the size of the image. 
    Therefore the offset from the image origin must be given in integer form to describe its placement.
    A layer contains one or more channels. 
    """

    def __init__(self, name):
        self.name     = name
        self.offset   = [0,0]
        self.dims     = [0,0]
        self.channels = {}

    def set_offset(self, x, y):
        self.offset = [x, y]

    def set_dims(self, w, h):
        self.dims = [w, h]

    def add_channel(self, name):
        self.channels[name] = channel.channel(name)
        self.channels[name].set_dims(self.dims[0], self.dims[1])

        return self.channels[name]

    def get_channels(self):
        for c in self.channels:
            yield c

    def get_channel(self, name):
        result = False
        if name in self.channels:
            result = self.channels[name]
        return result



import click
import json
import logging
from pathlib import Path

from tomviz import executor

from .load_data_source import create_read_options, extract_data_path
from .data_source_dependencies import load_dependencies

logger = logging.getLogger('tomviz')


def _create_data_source_options(data_sources):
    ids = [x['id'] for x in data_sources]
    filenames = [x['reader']['fileNames'][0] for x in data_sources]
    return dict(zip(ids, filenames))


def _data_source_options_string(options):
    s = ''
    for k, v in options.items():
        s += f'{k}  ({v})\n\n'

    return s


def _create_data_source_options_string(data_sources):
    options = _create_data_source_options(data_sources)
    return _data_source_options_string(options)


def _auto_select_data_source(data_sources):
    if len(data_sources) == 1:
        # Only one data source. Select this one.
        if 'id' not in data_sources[0]:
            # For backwards compatibility, when we did not have data
            # source ids (as is being done in the CI), make an id.
            data_sources[0]['id'] = '0x0'
        return data_sources[0]['id']

    with_pipeline = [len(x.get('operators', [])) > 0 for x in data_sources]
    if sum(with_pipeline) == 0:
        raise Exception('No operators were found')

    if sum(with_pipeline) > 1:
        options = _create_data_source_options_string(data_sources)
        msg = (
            'Multiple data sources with operators were found. One must be '
            'selected explicitly with the `-x` argument so we know which '
            f'pipeline to use. Available options:\n\n{options}'
        )
        raise Exception(msg)

    return data_sources[with_pipeline.index(True)]['id']


def _extract_pipeline(state, selected_data_source=None):
    if 'dataSources' not in state:
        raise Exception('Invalid state file: \'dataSources\' not found.')

    data_sources = state['dataSources']

    if not data_sources:
        raise Exception('No data source found.')

    if selected_data_source is None:
        selected_data_source = _auto_select_data_source(data_sources)

    ids = [x['id'] for x in data_sources]
    if selected_data_source not in ids:
        options = _create_data_source_options_string(data_sources)
        msg = (
            f'{selected_data_source} is not in the list of data sources! '
            f'Available options:\n\n{options}'
        )
        raise Exception(msg)

    data_source = data_sources[ids.index(selected_data_source)]

    if 'operators'not in data_source:
        raise Exception('\'operators\' not found.')

    operators = data_source['operators']

    if len(operators) == 0:
        raise Exception('No operators found.')

    return data_source, operators


@click.command(name="tomviz")
@click.option('-d', '--data-path', help='Path to an EMD/Data Exchange file or'
              ' directory containing EMD/Data Exchange files, can be used'
              ' to override data source in state file. If multiple files are'
              ' provided the pipeline with be run for each file.',
              type=click.Path(exists=True))
@click.option('-s', '--state-file-path', help='Path to the Tomviz state file',
              type=click.Path(exists=True), required=True)
@click.option('-x', '--selected-data-source',
              help='Explicitly select the id of the data source from which '
                   'the pipeline will be used',
              type=str, default=None)
@click.option('-o', '--output-file-path',
              help='Path to write the transformed dataset.', type=click.Path())
@click.option('-p', '--progress-method',
              help='The method to use to progress updates.',
              type=click.Choice(['tqdm', 'socket', 'files']), default='tqdm')
@click.option('-u', '--socket-path',
              help='The socket path to use for progress updates.',
              type=click.Path(), default='/tomviz/progress')
@click.option('-i', '--operator-index',
              help='The operator to start at.',
              type=int, default=0)
def main(data_path, state_file_path, output_file_path, progress_method,
         socket_path, operator_index, selected_data_source):

    # Extract the pipeline
    with open(state_file_path) as fp:
        state = json.load(fp, encoding='utf-8')

    data_source, operators = _extract_pipeline(state, selected_data_source)

    # Load the dependencies for the data source
    dependencies = load_dependencies(data_source, state, state_file_path,
                                     progress_method, socket_path)

    # Create the read options
    read_options = create_read_options(data_source)

    # if we have been provided a data file path we are going to use the one
    # from the state file, so check it exists.
    if data_path is None:
        data_path = extract_data_path(data_source, state_file_path)

    exts = ['.emd', '.h5', '.hdf5']
    data_path = Path(data_path)
    # Do we have multiple files to operate on
    if data_path.is_dir():
        data_file_paths = []
        for ext in exts:
            data_file_paths += list(data_path.glob('*' + ext))

        output_file_paths \
            = ['%s_transformed.emd' % x.stem for x in data_file_paths]
        # If we have been give an output directory write the output there
        if output_file_path is not None and Path(output_file_path).is_dir():
            output_file_paths \
                = [Path(output_file_path) / x for x in output_file_paths]
    else:
        if data_path.suffix.lower() not in exts:
            raise Exception(
                'Unsupported data source format, only HDF5 formats supported.')
        data_file_paths = [data_path]
        output_file_paths = [output_file_path]

    number_of_files = len(data_file_paths)
    if number_of_files == 0:
        logger.warning('No data files found.')
    elif number_of_files > 1:
        logger.info('Executing pipeline on %d files.' % number_of_files)

    for (data_file_path, output_file_path) in zip(data_file_paths,
                                                  output_file_paths):
        logger.info('Executing pipeline on %s' % data_file_path)
        executor.execute(operators, operator_index, data_file_path,
                         output_file_path, progress_method, socket_path,
                         read_options, dependencies)



import tomviz.operators


class BinaryMinMaxCurvatureFlow(tomviz.operators.CancelableOperator):

    def transform(self, dataset, stencil_radius=2, iterations=10,
                  threshold=50.0):
        """This filter smooths a binary image by evolving a level set with a
        curvature-based speed function. The Stencil Radius determines the scale
        of the noise to remove. The Threshold determines the iso-contour
        brightness to discriminate between two pixel classes.
        """

        # Initial progress
        self.progress.value = 0
        self.progress.maximum = 100

        # Approximate percentage of work completed after each step in the
        # transform
        STEP_PCT = [10, 20, 30, 70, 80, 90, 100]

        try:
            import itk
            import itkTypes
            from tomviz import itkutils
        except Exception as exc:
            print("Could not import necessary module(s)")
            raise exc

        # Return values
        returnValue = None

        # Add a try/except around the ITK portion. ITK exceptions are
        # passed up to the Python layer, so we can at least report what
        # went wrong with the script, e.g,, unsupported image type.
        try:
            self.progress.value = STEP_PCT[0]
            self.progress.message = "Converting data to ITK image"
            # Get the ITK image
            itk_image = itkutils.dataset_to_itk_image(dataset)
            itk_input_image_type = type(itk_image)
            self.progress.message = "Casting input to float type"
            itk_filter_image_type = itk.Image[itkTypes.F,
                                              itk_image.GetImageDimension()]
            caster = itk.CastImageFilter[itk_input_image_type,
                                         itk_filter_image_type].New()
            caster.SetInput(itk_image)
            itkutils.observe_filter_progress(self, caster,
                                             STEP_PCT[1], STEP_PCT[2])

            try:
                caster.Update()
            except RuntimeError:
                return

            self.progress.message = "Running filter"
            smoothing_filter = itk.BinaryMinMaxCurvatureFlowImageFilter[
                itk_filter_image_type, itk_filter_image_type].New()
            smoothing_filter.SetThreshold(threshold)
            smoothing_filter.SetStencilRadius(stencil_radius)
            smoothing_filter.SetNumberOfIterations(iterations)
            smoothing_filter.SetTimeStep(0.0625)
            smoothing_filter.SetInput(caster)
            itkutils.observe_filter_progress(self, smoothing_filter,
                                             STEP_PCT[2], STEP_PCT[3])

            try:
                smoothing_filter.Update()
            except RuntimeError:
                return

            itk_image_data = smoothing_filter.GetOutput()

            # Cast output to the input type
            self.progress.message = "Casting output to input type"

            caster = itk.CastImageFilter[itk_filter_image_type,
                                         itk_input_image_type].New()
            caster.SetInput(itk_image_data)
            itkutils.observe_filter_progress(self, caster,
                                             STEP_PCT[3], STEP_PCT[4])

            try:
                caster.Update()
            except RuntimeError:
                return

            itk_image_data = caster.GetOutput()

            self.progress.value = STEP_PCT[5]
            self.progress.message = "Saving results"

            itkutils.set_itk_image_on_dataset(itk_image_data, dataset)

            self.progress.value = STEP_PCT[6]

        except Exception as exc:
            print("Problem encountered while running %s" %
                  self.__class__.__name__)
            raise exc

        return returnValue



import numpy as np
from numpy.fft import fftn, fftshift, ifftn, ifftshift
import tomviz.operators


class ArtifactsTVOperator(tomviz.operators.CancelableOperator):

    def transform(self, dataset, Niter=100, a=0.1, wedgeSize=5, kmin=5,
                  theta=0):
        """
        Remove Structured Artifacts with Total Variation Minimization"""

        #Import information from dataset
        array = dataset.active_scalars
        (nx, ny, nz) = array.shape

        # Convert angle from Degrees to Radians.
        theta = (theta+90)*(np.pi/180)
        dtheta = wedgeSize*(np.pi/180)

        #Create coordinate grid in polar
        x = np.arange(-nx/2, nx/2-1, dtype=np.float64)
        y = np.arange(-ny/2, ny/2-1, dtype=np.float64)
        [x, y] = np.meshgrid(x, y, indexing='ij')
        rr = (np.square(x) + np.square(y))
        phi = np.arctan2(x, y)

        #Create the Angular Mask
        mask = np.ones((nx, ny), dtype=np.int8)
        mask[np.where((phi >= (theta-dtheta/2)) &
                      (phi <= (theta+dtheta/2)))] = 0
        mask[np.where((phi >= (np.pi+theta-dtheta/2)) &
                      (phi <= (np.pi+theta+dtheta/2)))] = 0
        mask[np.where((phi >= (-np.pi+theta-dtheta/2)) &
                      (phi <= (-np.pi+theta+dtheta/2)))] = 0
        mask[np.where(rr < np.square(kmin))] = 1 # Keep values below rmin.
        mask = np.array(mask, dtype=bool)

        # Initialize Progress bar.
        self.progress.maximum = nz * Niter

        #Main Loop
        for i in range(nz):

            #FFT of the Original Image.
            FFT_image = fftshift(fftn(array[:, :, i]))

            # Reconstruction starts as random image.
            recon_init = np.random.rand(nx, ny)

            self.progress.message = 'Processing Image No.%d/%d ' % (i+1, nz)

            #TV Artifact Removal Loop
            for j in range(Niter):

                # FFT of Reconstructed Image.
                FFT_recon = fftshift(fftn(recon_init))

                # Impose the Data Constraint
                FFT_recon[mask] = FFT_image[mask]

                #Inverse FFT
                recon_constraint = np.real(ifftn(ifftshift(FFT_recon)))

                # Positivity Constraint
                recon_constraint[recon_constraint < 0] = 0

                # TV Minimization Loop
                recon_minTV = recon_constraint
                d = np.linalg.norm(recon_minTV - recon_init)
                for k in range(20):
                    vst = TVDerivative(recon_minTV, nx, ny)
                    recon_minTV = recon_minTV - a*d*vst

                    if self.canceled:
                        return

                # Initializte the Next Loop.
                recon_init = recon_minTV

                # Update the Progress Bar.
                self.progress.value = i*Niter + j

            # Return reconstruction into stack.
            array[:, :, i] = recon_constraint

        #Set the result as the new scalars.
        dataset.active_scalars = np.asfortranarray(array)


def TVDerivative(img, nx, ny):
    fxy = np.pad(img, (1, 1), 'constant', constant_values=0)
    fxnegy = np.roll(fxy, -1, axis=0) #up
    fxposy = np.roll(fxy, 1, axis=0)  #down
    fnegxy = np.roll(fxy, -1, axis=1) #left
    fposxy = np.roll(fxy, 1, axis=1)  #right
    fposxnegy = np.roll(np.roll(fxy, 1, axis=1), -1, axis=0) #right and up
    fnegxposy = np.roll(np.roll(fxy, -1, axis=1), 1, axis=0) #left and down
    vst1 = (2*(fxy - fnegxy) +
            2*(fxy - fxnegy))/(np.sqrt(1e-8 +
                                       (fxy - fnegxy)**2 + (fxy - fxnegy)**2))
    vst2 = (2*(fposxy -
               fxy))/np.sqrt(1e-8 +
                             (fposxy - fxy)**2 + (fposxy - fposxnegy)**2)
    vst3 = (2*(fxposy -
               fxy))/np.sqrt(1e-8 +
                             (fxposy - fxy)**2 + (fxposy - fnegxposy)**2)
    vst = vst1 - vst2 - vst3
    vst = vst[1:-1, 1:-1]
    vst = vst/np.linalg.norm(vst)
    return vst



import logging
from pathlib import Path

from tomviz import executor

logger = logging.getLogger('tomviz')


def create_read_options(data_source):
    read_options = {}
    if 'subsampleSettings' in data_source:
        key = 'subsampleSettings'
        read_options[key] = data_source[key].copy()
    if 'keepCOrdering' in data_source:
        read_options['keep_c_ordering'] = data_source['keepCOrdering']

    return read_options


def extract_data_path(data_source, state_file_path):
    if 'reader' not in data_source:
        raise Exception('Data source does not contain a reader.')
    filenames = data_source['reader']['fileNames']
    if len(filenames) > 1:
        raise Exception('Image stacks not supported.')
    data_path = filenames[0]
    # fileName is relative to the state file location, so convert to
    # absolute path.
    data_path = (Path(state_file_path).parent / data_path).resolve()
    if not data_path.exists():
        raise Exception('Data source path does not exist: %s'
                        % data_path)

    return data_path


def load_data_source(data_source, state_file_path):
    read_options = create_read_options(data_source)
    data_path = extract_data_path(data_source, state_file_path)

    logger.info(f'Loading data at: {data_path}')
    return executor.load_dataset(data_path, read_options)



r"""Module used to generate Catalyst export scripts"""
from .. import simple, smstate, smtrace, servermanager

def _get_catalyst_state(options):
    # build a `source_set` comprising of the extractor proxies.
    # if not extracts have been configured, then there's nothing to generate.
    extractors = simple.GetExtractors()
    if not extractors:
        return None
    # convert catalyst options to PythonStateOptions.
    soptions = servermanager.ProxyManager().NewProxy("pythontracing", "PythonStateOptions")
    soptions = servermanager._getPyProxy(soptions)
    soptions.PropertiesToTraceOnCreate = smstate.RECORD_MODIFIED_PROPERTIES
    soptions.SkipHiddenDisplayProperties = True
    soptions.SkipRenderingComponents = False
    soptions.ExtractsOutputDirectory = options.ExtractsOutputDirectory
    return smstate.get_state(options=soptions, source_set=extractors,
            preamble=_get_catalyst_preamble(options),
            postamble=_get_catalyst_postamble(options))

def _get_catalyst_preamble(options):
    """returns the preamble text"""
    return ["# script-version: 2.0",
            "# Catalyst state generated using %s" % simple.GetParaViewSourceVersion()]

def _get_catalyst_postamble(options):
    """returns the postamble text"""
    trace_config = smtrace.start_trace(preamble="")
    trace_config.SetFullyTraceSupplementalProxies(True)

    # flush out some of the header since its not applicable here.
    smtrace.get_current_trace_output_and_reset()

    trace = smtrace.TraceOutput()
    trace.append_separated([\
        "# " + "-"*78,
        '# Catalyst options',
        "from paraview import catalyst"])
    accessor = smtrace.ProxyAccessor("options", options)
    trace.append(accessor.trace_ctor("catalyst.Options",
        smtrace.ProxyFilter()))
    del accessor
    smtrace.stop_trace()
    del trace_config
    trace.append_separated([\
        "# " + "-"*78,
        "if __name__ == '__main__':",
        "    from paraview.simple import SaveExtractsUsingCatalystOptions",
        "    # Code for non in-situ environments; if executing in post-processing",
        "    # i.e. non-Catalyst mode, let's generate extracts using Catalyst options",
        "    SaveExtractsUsingCatalystOptions(options)"])
    return str(trace)

def save_catalyst_state(fname, options):
    options = servermanager._getPyProxy(options)
    state = _get_catalyst_state(options)
    if not state:
        raise RuntimeError("No state generated")

    with open(fname, 'w') as file:
        file.write(state)
        file.write('\n')



r"""
Module that creates a source to match the data generated by
`paraview.demos.wavelet_miniapp`

This module supports the ability to be loaded in as a plugin in ParaView.
When loaded as a plugin, this adds a new data source named "Wavelet Mini-App"
(WaveletMiniApp) that can be used to produce data similar to when the
`wavelet_miniapp` is executed with properties matching the command line arguments
accepted by that module.

This is useful for generating tests or states that can be used with the miniapp.
"""

from paraview.demos.wavelet_miniapp import parser, create_dataset

#----------------------------------------------------------------
# If this module is loaded as a Plugin in ParaView, we create a source
# that acts similar to the miniapp.This is useful for testing.
from paraview.util.vtkAlgorithm import *
@smproxy.source(name="WaveletMiniApp",
                label="Wavelet Mini-App")
class WaveletMiniApp(VTKPythonAlgorithmBase):
    def __init__(self):
        VTKPythonAlgorithmBase.__init__(self, nInputPorts=0, nOutputPorts=1, outputType='vtkImageData')

        global parser
        self._args = parser.parse_args(["-s","not-used"])

    def _get_update_time(self, outInfo):
        executive = self.GetExecutive()
        timesteps = self.GetTimestepValues()
        if timesteps is None or len(timesteps) == 0:
            return None
        elif outInfo.Has(executive.UPDATE_TIME_STEP()) and len(timesteps) > 0:
            import bisect
            utime = outInfo.Get(executive.UPDATE_TIME_STEP())
            i = bisect.bisect_right(timesteps, utime)
            if i:
                return timesteps[i-1]
            return timesteps[0]
        else:
            assert(len(timesteps) > 0)
            return timesteps[0]

    def _get_update_piece(self, outInfo):
        executive = self.GetExecutive()
        return (executive.GetUpdatePiece(outInfo), executive.GetUpdateNumberOfPieces(outInfo))

    @smproperty.intvector(name="NumberOfTimeSteps", default_values=100)
    @smdomain.intrange(min="1")
    def SetTimeSteps(self, count):
        """Specify the number of timesteps (>=1)"""
        if self._args.timesteps != count:
            self._args.timesteps = count
            self.Modified()

    @smproperty.intvector(name="Size", default_values=101)
    @smdomain.intrange(min="1")
    def SetSize(self, size):
        """Specify the size (>=1) for the output datasets"""
        if self._args.size != size:
            self._args.size = size
            self.Modified()

    @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
    def GetTimestepValues(self):
        return [step/float(self._args.timesteps) for step in range(self._args.timesteps)]

    def RequestInformation(self, request, inInfoVec, outInfoVec):
        executive = self.GetExecutive()
        outInfo = outInfoVec.GetInformationObject(0)
        outInfo.Remove(executive.TIME_STEPS())
        outInfo.Remove(executive.TIME_RANGE())
        timesteps = self.GetTimestepValues()
        if timesteps is not None:
            for t in timesteps:
                outInfo.Append(executive.TIME_STEPS(), t)
            outInfo.Append(executive.TIME_RANGE(), timesteps[0])
            outInfo.Append(executive.TIME_RANGE(), timesteps[-1])

        ext = (self._args.size - 1) // 2
        outInfo.Set(executive.WHOLE_EXTENT(), -ext, ext, -ext, ext, -ext, ext)

        from vtkmodules.vtkCommonDataModel import vtkDataObject
        outInfo.Set(vtkDataObject.SPACING(), 1, 1, 1)
        return 1

    def RequestData(self, request, inInfoVec, outInfoVec):
        from vtkmodules.vtkCommonDataModel import vtkDataObject, vtkStreamingDemandDrivenPipeline
        data_time = self._get_update_time(outInfoVec.GetInformationObject(0))
        timestep = self.GetTimestepValues().index(data_time)
        piece, num_pieces = self._get_update_piece(outInfoVec.GetInformationObject(0))
        dataset, wholeExtent = create_dataset(timestep, self._args, piece, num_pieces)
        output = vtkDataObject.GetData(outInfoVec, 0)
        output.ShallowCopy(dataset)
        output.GetInformation().Set(output.DATA_TIME_STEP(), data_time)
        return 1



"""
module for generating cinema databases from arbitrary codes.
Essentially this provides a way to runs over all possible combinations of
values in database and call out to arbitrary codes so that they have a chance
to make corresponding changes before inserting the result into the database.
"""
from __future__ import absolute_import

from ..database import store


class Explorer(object):
    """
    Middleman that connects an arbitrary producing codes to the CinemaStore.
    The purpose of this class is to run through the parameter sets, and tell a
    set of tracks (in order) to do something with the parameter values
    it cares about.
    """

    def __init__(
        self,
        store,
        parameters,  # the things that this explorer is responsible for
        tracks  # the things we pass off values to in order to do the work
    ):
        self.__store = store
        self.parameters = parameters
        self.tracks = tracks

    @property
    def store(self):
        return self.__store

    def list_parameters(self):
        """
        parameters is an ordered list of parameters that the Explorer varies
        over
        """
        return self.parameters

    def prepare(self):
        """ Give tracks a chance to get ready for a run """
        if self.tracks:
            for e in self.tracks:
                e.prepare(self)

    def execute(self, desc):
        # Create the document/data product for this sample.
        doc = store.Document(desc)
        for e in self.tracks:
            # print ("EXECUTING track ", e, doc.descriptor)
            e.execute(doc)
        self.insert(doc)

    def explore(self, fixedargs=None, progressObject=None):
        """
        Explore the problem space to populate the store being careful not to
        hit combinations where dependencies are not satisfied.
        Fixed arguments are the parameters that we want to hold constant in
        the exploration.
        """
        self.prepare()

        for descriptor in self.store.iterate(
                self.list_parameters(), fixedargs, progressObject):
            self.execute(descriptor)

        self.finish()

    def finish(self):
        """ Give tracks a chance to clean up after a run """
        if self.tracks:
            for e in self.tracks:
                e.finish()

    def insert(self, doc):
        self.store.insert(doc)


class Track(object):
    """
    abstract interface for things that can produce data

    to use this:
    caller should set up some visualization
    then tie a particular set of parameters to an action with a track
    """

    def __init__(self):
        pass

    def prepare(self, explorer):
        """ subclasses get ready to run here """
        pass

    def finish(self):
        """ subclasses cleanup after running here """
        pass

    def execute(self, document):
        """ subclasses operate on parameters here"""
        pass


class LayerControl(object):
    """
    Prototype for something that Layer track can control
    """
    def __init__(self, name, showFunc, hideFunc):
        self.name = name
        self.callShow = showFunc
        self.callHide = hideFunc


class Layer(Track):
    """
    A track that connects a layer to the set of objects in the scene that it
    controls.
    """
    def __init__(self, layer, objectlist):
        super(Layer, self).__init__()
        self.parameter = layer
        # objlist is an array of class instances, they must have a name and
        # show and hide method, use LayerControl to make them.
        self.objectlist = objectlist

    def execute(self, doc):
        o = None
        if self.parameter in doc.descriptor:
            o = doc.descriptor[self.parameter]
        for obj in self.objectlist:
            if obj.name == o:
                try:
                    obj.callShow()  # method
                except TypeError:
                    obj.callShow(obj)  # function
            else:
                try:
                    obj.callHide()
                except TypeError:
                    obj.callHide(obj)



r"""Module used by vtkCDBWriter"""
from ..tpl import cinemasci

def write(fname, vtktable):
    cdb = cinemasci.new("cdb", { "path" : fname})
    if not cdb.initialize(dirExistCheck=False):
        raise RuntimeError("Failed to initialize dbase")
    for row in range(vtktable.GetNumberOfRows()):
        entry = {}
        for col in range(vtktable.GetNumberOfColumns()):
            value = vtktable.GetValue(row, col).ToString()
            if value:
                entry[vtktable.GetColumnName(col)] = value
        cdb.add_entry(entry)
    cdb.finalize()
    return True



import h5py
import numpy

class Reader:
    """ A HDF5 CIS Reader. """
    def __init__(self):
        return

    def read(self, cis):
        with h5py.File(cis.fname, "r") as f:
            cis.classname  = f.attrs["class"]
            cis.dims       = numpy.array(list(map(int, f.attrs["dims"].split(","))))
            cis.version    = f.attrs["version"]
            cis.flags      = f.attrs["flags"]

            for i in f["image"]:
                im = cis.add_image(i)
                self.read_image(im, f["image"][i])

    
    def read_image(self, image, data):
        for l in data["layer"]:
            layer = image.add_layer(l)
            # print("  l:{}".format(l))
            self.read_layer(layer, data["layer"][l])
        return

    def read_layer(self, layer, data):
        if "offset" in data.attrs.keys():
            offset = numpy.array(list(map(int, data.attrs["offset"].split(","))))
            layer.set_offset(offset[0], offset[1])
        if "dims" in data.attrs.keys():
            dims = numpy.array(list(map(int, data.attrs["dims"].split(","))))
            layer.set_dims(dims[0], dims[1])

        for c in data["channel"]:
            channel = layer.add_channel(c)
            # print("    c:{}".format(c))
        return



r"""This is a module to perform live visualization in the context of Catalyst.
"""

from paraview.simple import *
from paraview import servermanager

# -----------------------------------------------------------------------------
def ConnectToCatalyst(ds_host='localhost', ds_port=22222):
    """Creates a connection to a catalyst simulation. Example usage:

    > Connect("amber") # Connect to a single server at default port
    > Connect("amber", 12345) # Connect to a single server at port 12345

    return a LiveInsituLink object.
    """
    displaySession = servermanager.ActiveConnection.Session
    insituLink = servermanager.ConnectToCatalyst(ds_host, ds_port)
    ProcessServerNotifications()
    insituLink.InvokeCommand('Initialize')
    SetActiveConnection(servermanager.GetConnectionFromSession(displaySession))
    return insituLink

# -----------------------------------------------------------------------------
def ExtractCatalystData(link, name):
    ''' Extract data called "name" from simulation managed by the LiveInsituLink link.
        Register it in displaySession
    '''
    displaySession = servermanager.ActiveConnection.Session
    proxy = link.CreateExtract('sources', name, 0)
    displaySession.GetSessionProxyManager().RegisterProxy('sources', name, proxy)
    data = FindSource(name)
    return data

# -----------------------------------------------------------------------------
def PauseCatalyst(link, pause = True):
    ''' Pause / Unpause the Catalyst simulation managed by the LiveInsituLink link.
    '''
    currentState = link.GetProperty('SimulationPaused').GetElement(0) == 1
    if currentState == pause:
        return

    link.GetProperty('SimulationPaused').SetElement(0, 1 if pause else 0)
    link.UpdateVTKObjects()
    if not pause:
        link.InvokeCommand('LiveChanged')

# -----------------------------------------------------------------------------
def ProcessServerNotifications():
    ''' Processes pending events from server.
    '''
    pm = servermanager.vtkProcessModule.GetProcessModule()
    nam = pm.GetNetworkAccessManager()
    while nam.ProcessEvents(1) == 1:
        pass



import tomviz.operators


class OtsuMultipleThreshold(tomviz.operators.CancelableOperator):

    def transform(self, dataset, number_of_thresholds=1,
                  enable_valley_emphasis=False):
        """This filter performs semi-automatic multithresholding of a data set.
        Voxels are automatically classified into a chosen number of classes such
        that inter-class variance of the voxel values is minimized. The output
        is a label map with one label per voxel class.
        """

        # Initial progress
        self.progress.value = 0
        self.progress.maximum = 100

        # Approximate percentage of work completed after each step in the
        # transform
        STEP_PCT = [10, 20, 70, 90, 100]

        try:
            import itk
            import itkExtras
            import itkTypes
            from tomviz import itkutils
        except Exception as exc:
            print("Could not import necessary module(s)")
            raise exc

        # Return values
        returnValues = None

        # Add a try/except around the ITK portion. ITK exceptions are
        # passed up to the Python layer, so we can at least report what
        # went wrong with the script, e.g,, unsupported image type.
        try:
            self.progress.value = STEP_PCT[0]
            self.progress.message = "Converting data to ITK image"

            # Get the ITK image
            itk_image = itkutils.dataset_to_itk_image(dataset)
            itk_input_image_type = type(itk_image)

            # OtsuMultipleThresholdsImageFilter's wrapping requires that the
            # input and output image types be the same.
            itk_threshold_image_type = itk_input_image_type

            # Otsu multiple threshold filter
            otsu_filter = itk.OtsuMultipleThresholdsImageFilter[
                itk_input_image_type, itk_threshold_image_type].New()
            otsu_filter.SetNumberOfThresholds(number_of_thresholds)
            otsu_filter.SetValleyEmphasis(enable_valley_emphasis)
            otsu_filter.SetInput(itk_image)
            itkutils.observe_filter_progress(self, otsu_filter,
                                             STEP_PCT[1], STEP_PCT[2])

            try:
                otsu_filter.Update()
            except RuntimeError:
                return

            print("Otsu threshold(s): %s" % (otsu_filter.GetThresholds(),))

            itk_image_data = otsu_filter.GetOutput()

            # Cast threshold output to an integral type if needed.
            py_buffer_type = itk_threshold_image_type
            voxel_type = itkExtras.template(itk_threshold_image_type)[1][0]
            if voxel_type is itkTypes.F or voxel_type is itkTypes.D:
                self.progress.message = "Casting output to integral type"

                # Unsigned char supports 256 labels, or 255 threshold levels.
                # This should be sufficient for all but the most unusual use
                # cases.
                py_buffer_type = itk.Image.UC3
                caster = itk.CastImageFilter[itk_threshold_image_type,
                                             py_buffer_type].New()
                caster.SetInput(itk_image_data)
                itkutils.observe_filter_progress(self, caster,
                                                 STEP_PCT[2], STEP_PCT[3])

                try:
                    caster.Update()
                except RuntimeError:
                    return

                itk_image_data = caster.GetOutput()

            self.progress.value = STEP_PCT[3]
            self.progress.message = "Saving results"

            label_map_dataset = dataset.create_child_dataset()
            itkutils.set_itk_image_on_dataset(itk_image_data, label_map_dataset,
                                              dtype=py_buffer_type)

            self.progress.value = STEP_PCT[4]

            # Set up dictionary to return operator results
            returnValues = {}
            returnValues["label_map"] = label_map_dataset

        except Exception as exc:
            print("Problem encountered while running %s" %
                  self.__class__.__name__)
            raise exc

        return returnValues



def transform(dataset):
    """
    Normalize tilt series so that each tilt image has the same total intensity.
    """

    import numpy as np

    data = dataset.active_scalars # Get data as numpy array

    if data is None: # Check if data exists
        raise RuntimeError("No data array found!")

    data = data.astype(np.float32) # Change tilt series type to float.

    # Calculate average intensity of tilt series.
    intensity = np.sum(np.average(data, 2))

    for i in range(0, data.shape[2]):
        # Normalize each tilt image.
        data[:, :, i] = data[:, :, i] / np.sum(data[:, :, i]) * intensity

    dataset.active_scalars = data



try:
    from .load_output import list_elements, extract_elements  # noqa
    from .make_hdf5 import make_hdf5  # noqa
    from .process_projections import fix_python_paths, ic_names, process_projections  # noqa
    requirements_installed = True
except ImportError:
    requirements_installed = False


def installed():
    return requirements_installed



from tomviz._wrapping import PipelineStateManagerBase


class PipelineStateManager(PipelineStateManagerBase):
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = PipelineStateManagerBase.__new__(cls, *args,
                                                             **kwargs)

        return cls._instance



# -*- coding: utf-8 -*-

###############################################################################
# This source file is part of the Tomviz project, https://tomviz.org/.
# It is released under the 3-Clause BSD License, see "LICENSE".
###############################################################################
import numpy as np

from tomviz.io import FileType, IOBase, Reader, Writer

import tomviz.utils

from vtk import vtkImageData


class NumpyBase(IOBase):

    @staticmethod
    def file_type():
        return FileType('NumPy binary format', ['npy'])


class NumpyWriter(Writer, NumpyBase):

    def write(self, path, data_object):
        data = tomviz.utils.get_array(data_object)

        # Convert to C ordering
        data = np.ascontiguousarray(data)

        with open(path, "wb") as f:
            np.save(f, data)


class NumpyReader(Reader, NumpyBase):

    def read(self, path):
        with open(path, "rb") as f:
            data = np.load(f)

        if len(data.shape) != 3:
            return vtkImageData()

        # Convert to Fortran ordering
        data = np.asfortranarray(data)

        image_data = vtkImageData()
        (x, y, z) = data.shape

        image_data.SetOrigin(0, 0, 0)
        image_data.SetSpacing(1, 1, 1)
        image_data.SetExtent(0, x - 1, 0, y - 1, 0, z - 1)
        tomviz.utils.set_array(image_data, data)

        return image_data



def transform(dataset, axis1, axis2):
    """Swap two axes in a dataset"""

    import numpy as np

    data_py = dataset.active_scalars
    swapped = data_py.swapaxes(axis1, axis2)
    dataset.active_scalars = np.asfortranarray(swapped)



def transform(dataset,  XRANGE=None, YRANGE=None, ZRANGE=None):
    """Define this method for Python operators that
    transform input scalars"""

    import numpy as np

    array = dataset.active_scalars
    if array is None:
        raise RuntimeError("No scalars found!")

    # Transform the dataset.
    result = np.copy(array)
    result[XRANGE[0]:XRANGE[1], YRANGE[0]:YRANGE[1], ZRANGE[0]:ZRANGE[1]] = 0
    # Set the result as the new scalars.
    dataset.active_scalars = result



"""
Creating images that correspond to a set of queries by compositing the
results together. Used by viewer primarily.
"""
from __future__ import absolute_import
import itertools as it

from .querymaker import QueryMaker
from . import layer_rasters
import copy


class QueryMaker_SpecB(QueryMaker):
    '''
    This class translates UI queries into a set of Spec-B LayerRasters
    instances (multi-image).  This set of LayerRasters can be composited
    through the Compositor class (compositor.py).
    '''
    def __init__(self):
        super(QueryMaker_SpecB, self).__init__()

    def _QueryMaker__createBaseLayerFromQuery(self, query):
        '''
        Creates a base LayerRasters containing the values that are shared
        (global) by the set of user-selected layers (e.g. current time,
        camera, etc.). This base LayerRasters will be used as an initial
        template for all of the LayerRasters.

        These queries are assumed to have a single value. Queries containing
        multiple values need to be added separately in __generateQueriedLayers
        (...).
        '''
        baseLayer = layer_rasters.LayerRasters()
        parameter_list = self.store().parameter_list

        for name in query.keys():
            if (name in parameter_list) and (
                    "role" not in parameter_list[name]):
                values = query[name]
                if name == 'pose':
                    v = list(iter(values))
                else:
                    # assumed to have a single value
                    v = list(iter(values))[0]
                baseLayer.addToBaseQuery({name: v})

        return baseLayer

    def __generateQueriedLayers(self, query, globalBaseLayer):
        '''
        Creates all the LayerRasters instances, uses baseLayer as a common
        template (__createLayer(..) creates copies of it).

        Iterates over the "vis" parameters (assumes only those available items
        are to be displayed) and creates a LayerRasters for each based on all
        its necessary parameters. This method heavily relies on
        cinema_store::parameters_for_object() in order to resolve all the
        necessary parameters.

        visQuery is the top-level query (vis parameter), parQueries are
        queries of the mid-level parent filters wich control values affect the
        final parameter node, ownQueries account for the control
        values of the parameter itself.
        '''
        createdLayers = []

        # every visible pipeline item
        for paramName in (query["vis"] if ("vis" in query) else []):
            relatedParams = self.store().parameters_for_object(paramName)
            fieldName = relatedParams[1]
            dependeeControls = relatedParams[2]
            visQuery = {"vis": paramName}

            # pop its own name from the dependee list, create specific queries
            ownQueries = []
            if paramName in dependeeControls:
                dependeeControls.pop(dependeeControls.index(paramName))
                ownQueries = self.__generateSpecificBaseQueries(
                    query, paramName)

            # add the field to the list for the case when there is no controls
            # but only field value arrays
            if len(ownQueries) == 0:
                ownQueries += self.__generateSpecificBaseQueries(
                    query, fieldName)

            # set up each of the parent control queries (shared by each
            # target query)
            parQueries = []
            for control in dependeeControls:
                # TODO: make the [] brackets default to __genSpecificQueries
                parQueries += [
                    self.__generateSpecificBaseQueries(query, control)]

            if len(parQueries) > 0:
                allQueries = self.__traverseQueryTree(parQueries, ownQueries)
            else:
                allQueries = ownQueries

            for targetq in allQueries:
                totalBaseQueries = [visQuery] + (
                    [targetq] if isinstance(targetq, dict) else list(targetq))
                layer = self.__createLayer(
                    query, globalBaseLayer, totalBaseQueries, fieldName)
                layer.setCustomizationName(paramName)
                createdLayers.append(layer)

        return createdLayers

    def __traverseQueryTree(self, parQueries, ownQueries):
        '''
        Assembles iterable objects with all of the controls parameter
        combinations (including both the parent control values and its own
        control values). Uses map to apply the my_list customized list
        initializer and itertools.chain.from_iterable to chain together each
        layer's list of parameters to obtain a final list with all
        of the chained combinations.

        parQueries (parent control queries):
            [[{'Ctrl1' : a}, {'Ctrl1' : b}, {'Ctrl1' : c}, ...],
             [{'Ctrl2' : a}, {'Ctrl2' : b}, {'Ctrl2' : c}, ...], ...]

        ownQueries (parameter's own control values):
            [{'param' : a}, {'param' : b}, {'param' : c}, ...]

        output (all of the combinations)**:

            [[{'Ctrl1' : a}, {'Ctrl2' : a}, {'param' : a}],
             [{'Ctrl1' : a}, {'Ctrl2' : a}, {'param' : b}],
             [{'Ctrl1' : a}, {'Ctrl2' : a}, {'param' : c}], ...]

        Note(**): The elements of the output list are itertool objects which
        need to be evaluated by initializing a list (or any other iterable).

        Warning: this function assumes ownQueries and parQueries contain
        at least one element.  The caller should ensure this is always true.

        Note: See itertools docs for more information.
        (https://docs.python.org/3/library/itertools.html#itertools.product)'''

        def my_list(iterable_var):
            '''
            This is a customized list initializer which gives special
            treatment to dict types (e.g. queries). All iterables are
            initialized with list() while dictionaries are appended to a list.
            The intention is to be able to pass this function to map() and
            obtain a list of lists which is very simple to chain
            together with itertools.chain.from_iterable().'''
            if isinstance(iterable_var, dict):
                return [iterable_var]
            else:
                return list(iterable_var)

        # Compute the combinations between parent controls. it.product gives
        # the inner product between the two lists and returns sets which are
        # then converted back to lists. The resulting list-elements are
        # concatenated into a single list.
        current = map(my_list, parQueries.pop(0))
        for nextQuery in parQueries:
            current = map(
                my_list, it.product(current, map(my_list, nextQuery)))
            current = map(it.chain.from_iterable, current)
            # Force it.chain evaluation to avoid recursive referencing to a
            # single generator when doing the last product (ln. 170), as gens
            # evaluate only once.
            current = map(my_list, current)

        # Compute the combinations of the parent controls with the
        # parameter's own control values
        allQueries = map(
            my_list, it.product(current, map(my_list, ownQueries)))
        allQueries = map(it.chain.from_iterable, allQueries)

        return allQueries

    def __generateSpecificBaseQueries(self, query, control):
        '''
        Creates a set of queries based on its input parameters. These are all
        specific to a particular LayerRasters or group of LayerRasters.
        '''
        return [{control: value} for value in (
            query[control] if (control in query) else [])]

    def __createLayer(self, query, baseLayer, otherBaseQueries, fieldName):
        '''
        Creates a new layer by copying a 'baseLayer' and then including
        additional base queries ('otherBaseQueries'). It then adds all the fbo
        components specified in a field parameter ('fieldName') to this layer
        for later compositing.
        '''
        newLayer = copy.deepcopy(baseLayer)
        for bq in otherBaseQueries:
            newLayer.addToBaseQuery(bq)

        try:
            fieldParameter = self.store().parameter_list[fieldName]
            fieldValues = fieldParameter["values"]
            fieldTypes = fieldParameter["types"]
            fieldRanges = fieldParameter["valueRanges"] if (
                "valueRanges" in fieldParameter) else {}

            for index, value in enumerate(fieldValues):

                # TODO: when "rgb" a hack might be needed to add a single
                # color when several available
                # Note 4th condition: includes a query[] value if the
                # fieldName has been queried (necessary
                # for compatibility SpecB-testcase1
                if (fieldTypes[index] == "depth" or
                    fieldTypes[index] == "luminance" or
                    fieldTypes[index] == "rgb" or
                    ((value in query[fieldName])
                     if (fieldName in query) else False)):

                    # Actual value ranges will be used to normalize color lut
                    # if set

                    # Ensure value rasters are float and not invertible
                    if "value_mode" in self.store().metadata:
                        self.__floatValueRasters = self.store().metadata[
                            "value_mode"] == 2
                    else:
                        self.__floatValueRasters = False

                    if (value in fieldRanges) and self.__floatValueRasters:
                        newLayer.setValueRange(fieldRanges[value])

                    imageType = self.store().determine_type({fieldName: value})
                    # TODO imageType parameter can be resolved  with the
                    # other two within ::addQuery()
                    newLayer.addQuery(imageType, fieldName, value)

        except KeyError:
            raise KeyError(
                "The field parameter ", fieldName, " is not well defined!",
                "\n Types: ", fieldTypes,
                "\n Values: ", fieldValues,
                "\n Query: ", query,
                "\n Value: ", value, " INDEX: ", index)

        return newLayer

    def translateQuery(self, query, colorDefinitions={}):
        '''
        Receives a UI query and returns a set of LayerRasters consisting of all
        the different layers to composite. See the base class documentation
        for detail.
        '''
        # Create a common base LayerRasters (time and camera).
        baseLayer = self._QueryMaker__createBaseLayerFromQuery(query)
        # Create all the LayerRasters (TODO cache layers and other db info to
        # avoid re-creating them every time).
        allLayers = self.__generateQueriedLayers(query, baseLayer)
        self._QueryMaker__loadLayers(allLayers)

        return allLayers

    def supportsLayering(self):
        return True



from tomviz import utils
import numpy as np
import tomviz.operators


class CrossCorrelationAlignmentOperator(tomviz.operators.CancelableOperator):

    def transform(self, dataset):
        """Automatically align tilt images by cross-correlation"""
        self.progress.maximum = 1

        tiltSeries = dataset.active_scalars.astype(float)
        tiltAngles = dataset.tilt_angles

        # determine reference image index
        zeroDegreeTiltImage = None
        if tiltAngles is not None:
            zeroDegreeTiltImage = np.where(tiltAngles == 0)[0]

        if zeroDegreeTiltImage:
            referenceIndex = zeroDegreeTiltImage[0]
        else:
            referenceIndex = tiltSeries.shape[2] // 2

        # create Fourier space filter
        filterCutoff = 4
        (Ny, Nx, Nproj) = tiltSeries.shape
        ky = np.fft.fftfreq(Ny)
        kx = np.fft.fftfreq(Nx)
        [kX, kY] = np.meshgrid(kx, ky)
        kR = np.sqrt(kX**2 + kY**2)
        kFilter = (kR <= (0.5 / filterCutoff)) * \
            np.sin(2 * filterCutoff * np.pi * kR)**2

        # create real sapce filter to remove edge discontinuities
        y = np.linspace(1, Ny, Ny)
        x = np.linspace(1, Nx, Nx)
        [X, Y] = np.meshgrid(x, y)
        rFilter = (np.sin(np.pi * X / Nx) * np.sin(np.pi * Y / Ny)) ** 2

        self.progress.maximum = tiltSeries.shape[2] - 1
        step = 1

        offsets = np.zeros((tiltSeries.shape[2], 2))

        for i in range(referenceIndex, Nproj - 1):
            if self.canceled:
                return
            self.progress.message = 'Processing tilt image No.%d/%d' % (
                step, Nproj)

            offsets[i + 1, :], tiltSeries[:, :, i + 1] = crossCorrelationAlign(
                tiltSeries[:, :, i + 1], tiltSeries[:, :, i], rFilter, kFilter)
            step += 1
            self.progress.value = step

        for i in range(referenceIndex, 0, -1):
            if self.canceled:
                return
            self.progress.message = 'Processing tilt image No.%d/%d' % (
                step, Nproj)
            offsets[i - 1, :], tiltSeries[:, :, i - 1] = crossCorrelationAlign(
                tiltSeries[:, :, i - 1], tiltSeries[:, :, i], rFilter, kFilter)
            step += 1
            self.progress.value = step

        dataset.active_scalars = tiltSeries

        # Assign Negative Shifts when Shift > N/2.
        indices_X = np.where(offsets[:, 0] > tiltSeries.shape[0] / 2)
        offsets[indices_X, 0] -= tiltSeries.shape[0]
        indices_Y = np.where(offsets[:, 1] > tiltSeries.shape[1] / 2)
        offsets[indices_Y, 1] -= tiltSeries.shape[1]

        # Create a spreadsheet data set from table data
        column_names = ["X Offset", "Y Offset"]
        offsetsTable = utils.make_spreadsheet(column_names, offsets)
        # Set up dictionary to return operator results
        returnValues = {}
        returnValues["alignments"] = offsetsTable
        return returnValues


def crossCorrelationAlign(image, reference, rFilter, kFilter):
    """Align image to reference by cross-correlation"""

    image_f = np.fft.fft2((image - np.mean(image)) * rFilter)
    reference_f = np.fft.fft2((reference - np.mean(reference)) * rFilter)

    xcor = abs(np.fft.ifft2(np.conj(image_f) * reference_f * kFilter))
    shifts = np.unravel_index(xcor.argmax(), xcor.shape)

    # shift image
    output = np.roll(image, shifts[0], axis=0)
    output = np.roll(output, shifts[1], axis=1)

    return shifts, output



from jsonpointer import resolve_pointer


def find_pipeline(path):
    from . import pipelines

    if len(path) < 2:
        raise ValueError("Path doesn't have enough parts.")

    obj_type = path.pop(0)
    if obj_type != 'dataSources':
        raise ValueError("Path doesn't start with 'dataSources'.")

    pipeline_index = int(path.pop(0))

    if pipeline_index >= len(pipelines):
        raise ValueError('Pipeline index no longer exists.')

    return pipelines[pipeline_index]


def find_operator(path):
    pipeline = find_pipeline(path)

    if len(path) < 1:
        raise ValueError("Path doesn't have enough parts.")

    obj_type = path.pop(0)
    if obj_type != "operators":
        raise ValueError("Path doesn't contain 'operators'.")

    op_index = -1
    if path:
        op_index = path.pop(0)
        op_index = int(op_index)

    operators = pipeline.dataSource.operators
    if op_index >= len(operators):
        raise ValueError("Operator index no longer exists.")

    return operators[op_index]


def find_datasource(path):
    copy = path.copy()
    pipeline = find_pipeline(copy)

    # If the path has no other dataSources we are done
    if 'dataSources' not in copy:
        del path[:2]

        return pipeline.dataSource

    # Find the operator (the child data source must be associated with one ...)
    op = find_operator(path)
    del path[:2]

    return op.dataSources[0]


def find_module(path):
    # First find the data source the module is attached to.
    datasource = find_datasource(path)
    if datasource is None:
        raise ValueError('Unable to find data source.')

    if len(path) < 1:
        raise ValueError("Path doesn't have enough parts.")

    obj_type = path.pop(0)
    if obj_type != 'modules':
        raise ValueError("Path doesn't contain 'modules'.")

    mod_index = -1
    if path:
        mod_index = path.pop(0)
        mod_index = int(mod_index)

    modules = datasource.modules
    if mod_index >= len(modules):
        raise ValueError("Module index no longer exists.")

    return modules[mod_index]


# Note: This can only use using during a _update_state(...) call!
def find_removed_object_id(path):
    # _state is the old state containing the remove object
    from . import _state
    obj = resolve_pointer(_state, path)

    return obj['id']


def _path(path, obj_type):
    path = path.split('/')
    # Find last occurrence
    index = len(path) - 1 - path[::-1].index(obj_type)
    path = path[0:index+2]

    return '/'.join(path)


def operator_path(path):
    return _path(path, 'operators')


def module_path(path):
    return _path(path, 'modules')


def datasource_path(path):
    return _path(path, 'dataSources')


def pipeline_index(path):
    return int(path.split('/')[2])



import tomviz.operators


class UnsharpMask(tomviz.operators.CancelableOperator):

    def transform(self, dataset, amount=0.5, threshold=0.0, sigma=1.0):
        """This filter performs anisotropic diffusion on an image using
        the classic Perona-Malik, gradient magnitude-based equation.
        """

        # Initial progress
        self.progress.value = 0
        self.progress.maximum = 100

        # Approximate percentage of work completed after each step in the
        # transform
        step_pct = iter([10, 20, 90, 100])

        try:
            import itk
            from tomviz import itkutils
        except Exception as exc:
            print("Could not import necessary module(s)")
            raise exc

        try:
            self.progress.message = "Converting data to ITK image"
            self.progress.value = 0

            # Get the ITK image.
            itk_image = itkutils.dataset_to_itk_image(dataset)
            self.progress.value = next(step_pct)

            self.progress.message = "Running filter"
            self.progress.value = next(step_pct)

            unsharp_mask = \
                itk.UnsharpMaskImageFilter.New(Input=itk_image)
            unsharp_mask.SetAmount(amount)
            unsharp_mask.SetThreshold(threshold)
            unsharp_mask.SetSigma(sigma)
            itkutils.observe_filter_progress(self, unsharp_mask,
                                             self.progress.value,
                                             next(step_pct))

            try:
                unsharp_mask.Update()
            except RuntimeError:
                return

            self.progress.message = "Saving results"

            itkutils.set_itk_image_on_dataset(unsharp_mask.GetOutput(), dataset)

            self.progress.value = next(step_pct)
        except Exception as exc:
            print("Problem encountered while running %s" %
                  self.__class__.__name__)
            raise exc



from tomviz import utils
import numpy as np
from scipy import ndimage
import tomviz.operators


class AutoTiltAxisRotationAlignOperator(tomviz.operators.CancelableOperator):

    def transform(self, dataset):
        """
          Automatic align the tilt axis to horizontal direction (x-axis)
        """
        self.progress.maximum = 1

        # Get Tilt Series
        tiltSeries = dataset.active_scalars

        if tiltSeries is None: #Check if data exists
            raise RuntimeError("No data array found!")

        (Nslice, Nray, Nproj) = tiltSeries.shape
        Intensity = np.zeros(tiltSeries.shape)

        self.progress.maximum = Nproj + 131 #coarse(90)+fine(41)
        step = 0

        self.progress.message = 'Initialization'
        #take FFT of all projections
        for i in range(Nproj):
            self.progress.message = ('Taking Fourier transform of tilt image'
                                     'No.%d/%d' % (i + 1, Nproj))
            tiltImage = tiltSeries[:, :, i]
            tiltImage_F = np.abs(np.fft.fft2(tiltImage))
            if (i == 0):
                temp = tiltImage_F[0, 0]
            Intensity[:, :, i] = np.fft.fftshift(tiltImage_F / temp)
            step += 1
            self.progress.value = step

        #rescale intensity
        Intensity = np.power(Intensity, 0.2)

        #calculate variation itensity image
        Intensity_var = np.var(Intensity, axis=2)
        #search angles
        coarseStep = 2
        fineStep = 0.1
        coarseAngles = np.arange(-90, 90, coarseStep)

        Nx = Intensity_var.shape[0]
        Ny = Intensity_var.shape[1]
        N = np.round(np.min([Nx, Ny]) // 3)

        #coarse search
        I = np.zeros((coarseAngles.size, N))
        for a in range(coarseAngles.size):
            if self.canceled:
                return
            self.progress.message = ('Calculating line intensity at angle %f'
                                     'degree' % (coarseAngles[a]))
            I[a, :] = calculateLineIntensity(Intensity_var, coarseAngles[a], N)
            step += 1
            self.progress.value = step

        I_sum = np.sum(I, axis=1)
        minIntensityIndex = np.argmin(I_sum)
        rot_ang = coarseAngles[minIntensityIndex]

        #now refine
        fineAngles = np.arange(rot_ang - coarseStep,
                               rot_ang + coarseStep + fineStep, fineStep)

        #fine search
        I = np.zeros((fineAngles.size, N))
        for a in range(fineAngles.size):
            if self.canceled:
                return
            self.progress.message = ('Calculating line intensity at angle %f'
                                     'degree' % (fineAngles[a]))

            I[a, :] = calculateLineIntensity(Intensity_var, fineAngles[a], N)
            step += 1
            self.progress.value = step

        I_sum = np.sum(I, axis=1)
        minIntensityIndex = np.argmin(I_sum)
        rot_ang = fineAngles[minIntensityIndex]

        self.progress.message = 'Rotating tilt series'
        axes = ((0, 1))
        shape = utils.rotate_shape(tiltSeries, -rot_ang, axes=axes)
        result = np.empty(shape, tiltSeries.dtype, order='F')
        ndimage.interpolation.rotate(
            tiltSeries, -rot_ang, axes=axes, output=result)

        print("rotate tilt series by %f degrees" % -rot_ang)

        # Set the result as the new scalars.
        dataset.active_scalars = result


def calculateLineIntensity(Intensity_var, angle_d, N):
    Nx = Intensity_var.shape[0]
    Ny = Intensity_var.shape[1]
    cenx = np.floor(Nx / 2)
    ceny = np.floor(Ny / 2)
    ang = angle_d * np.pi / 180
    w = np.zeros(N)
    v = np.zeros(N)
    for i in range(0, N):
        x = i * np.cos(ang)
        y = i * np.sin(ang)
        sx = abs(np.floor(x) - x)
        sy = abs(np.floor(y) - y)
        px = int(np.floor(x) + cenx)
        py = int(np.floor(y) + ceny)
        if (px >= 0 and px < Nx and py >= 0 and py < Ny):
            w[i] = w[i] + (1 - sx) * (1 - sy)
            v[i] = v[i] + (1 - sx) * (1 - sy) * Intensity_var[px, py]
        px = int(np.ceil(x) + cenx)
        py = int(np.floor(y) + ceny)
        if (px >= 0 and px < Nx and py >= 0 and py < Ny):
            w[i] = w[i] + sx * (1 - sy)
            v[i] = v[i] + sx * (1 - sy) * Intensity_var[px, py]
        px = int(np.floor(x) + cenx)
        py = int(np.ceil(y) + ceny)
        if (px >= 0 and px < Nx and py >= 0 and py < Ny):
            w[i] = w[i] + (1 - sx) * sy
            v[i] = v[i] + (1 - sx) * sy * Intensity_var[px, py]
        px = int(np.ceil(x) + cenx)
        py = int(np.ceil(y) + ceny)
        if (px >= 0 and px < Nx and py >= 0 and py < Ny):
            w[i] = w[i] + sx * sy
            v[i] = v[i] + sx * sy * Intensity_var[px, py]
    v[w != 0] = v[w != 0] / w[w != 0]
    return v



from tomviz.io import dm
from tomviz.io import ser
import numpy as np
import time
import os


# Class for listening and logging new files for real-time reconstruction
class Logger:

    def __init__(self, listenDirectory, fileExtension):

        # Create Local Directory
        if not os.path.exists(listenDirectory):
            os.makedirs(listenDirectory)

        # Set Member Variables
        self.listenDir = listenDirectory
        self.fileExt = fileExtension

        self.listenFiles = self.listen_files_list(self.listenDir)
        self.logFiles, self.logTiltSeries, self.logTiltAngles = [], [], []
        print("Listener on {} created.".format(self.listenDir))

    def listen_files_list(self, directory):
        """Grab current files in listen directory"""
        files = [f for f in os.listdir(directory) if
                 f[-len(self.fileExt):] == self.fileExt]
        return files

    def CHANGE_appendAll(self):
        """Append stored files for each new element"""
        # Separate new files to be loaded
        FoI = list(set(self.listenFiles)-set(self.logFiles))
        for file in FoI:
            print("Loading {}".format(file))
            filePath = os.path.join(self.listenDir, file)

            try:
                (newProj, newAngle) = self.read_projection_image(filePath)

                self.logTiltAngles = np.append(self.logTiltAngles, newAngle)

                newProj = self.background_subtract(newProj)
                newProj = self.center_of_mass_align(newProj)

                # Account for Python's disdain for AxAx1 arrays
                # (compresses to 2D)
                if(len(self.logTiltSeries) == 0):
                    dataDim = np.shape(newProj)
                    self.logTiltSeries = np.zeros([dataDim[0], dataDim[1], 1])
                    self.logTiltSeries[:, :, 0] = newProj
                else:
                    self.logTiltSeries = np.dstack((self.logTiltSeries,
                                                    newProj))

                self.logFiles = np.append(self.logFiles, file)

            except Exception:
                print('Could not read : {}, will proceed with reconstruction\
                        and re-download on next pass'.format(file))
                break

    def monitor(self, seconds=1):
        """Return true if, within seconds, any new files exist in listenDir
        NOTE: Lazy Scheme is used (set difference), can update """

        for ts in range(0, seconds):
            self.listenFiles = self.listen_files_list(self.listenDir)
            FoI = list(set(self.listenFiles)-set(self.logFiles))
            if len(FoI) == 0:
                time.sleep(1)
            else:
                self.CHANGE_appendAll() # Can be probamatic for first iter..
                return True

        return False

    def read_projection_image(self, fname):
        """Acquires angles from metadata of .dm4 files"""

        if self.fileExt == 'dm4':
            file = dm.FileDM(fname)
            alphaTag = '.ImageList.2.ImageTags.Microscope Info.'\
                       'Stage Position.Stage Alpha'
            return (file.getDataset(0)['data'], file.allTags[alphaTag])
        elif self.fileExt == 'ser':
            file = ser.SerReader(fname)
            return (file['data'], file['metadata']['Stage A [deg]'])
        # Stage Alpha isn't stored in metadata for dm3 or tif
        elif self.fileExt == 'dm3':
            # Parse fname for stage alpha
            file = dm.FileDM(fname)

            match = 'degrees'
            tags = fname.split('_')

            angle = [tag for tag in tags if match in tag][0]
            angleInd = angle.find(match)
            return (file.getDataset(0)['data'], float(angle[:angleInd]))

    def load_tilt_series(self, tomo, alg):

        if alg != 'WBP':
            (Nslice, Nray, Nproj) = self.logTiltSeries.shape
            b = np.zeros([Nslice, Nray * Nproj], dtype=np.float32)
            for s in range(Nslice):
                b[s, :] = self.logTiltSeries[s, :, :].transpose().ravel()
            tomo.set_tilt_series(b)
        else:
            tomo.set_tilt_series(self.logTiltSeries, self.logTiltAngles)

    # Shift Image so that the center of mass is at the origin"
    # Automatically align tilt images by center of mass method
    def center_of_mass_align(self, image):
        (Nx, Ny) = image.shape
        y = np.linspace(0, Ny-1, Ny)
        x = np.linspace(0, Nx-1, Nx)
        [X, Y] = np.meshgrid(x, y, indexing="ij")

        imageCOM_x = int(np.sum(image * X) / np.sum(image))
        imageCOM_y = int(np.sum(image * Y) / np.sum(image))

        sx = -(imageCOM_x - Nx // 2)
        sy = -(imageCOM_y - Ny // 2)

        output = np.roll(image, sx, axis=0)
        output = np.roll(output, sy, axis=1)

        return output

    # Remove Background Intensity
    def background_subtract(self, image):

        (Nx, Ny) = image.shape
        XRANGE = np.array([0, int(Nx//16)], dtype=int)
        YRANGE = np.array([0, int(Ny//16)], dtype=int)

        image -= np.average(image[XRANGE[0]:XRANGE[1], YRANGE[0]:YRANGE[1]])

        image[image < 0] = 0

        return image



import argparse, json, os, os.path, sys

def splitall(path):
    allparts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path: # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    return allparts


def find_subdirectory(subdir, node):
    contents = node["contents"]
    for item in contents:
        if item["type"] == "directory" and item["name"] == subdir:
            return item
    raise RuntimeError("Missing subdir '%s'" % subdir)

def find_directory(subdir, node):
    if subdir == '.' or subdir == None:
        return node
    for part in splitall(subdir):
        node = find_subdirectory(part, node)
    return node

def get_baseline(tname, path, image, root):
    import re
    filename = os.path.join(path, image)
    relpath = os.path.relpath(filename, root)
    return tname + "_" + re.sub('[^0-9a-zA-Z.]+', '_', relpath)

parser = argparse.ArgumentParser(\
        description="Validate Extrats generated in tests")

parser.add_argument("--name", type=str, required=True,
    help="name of the test being validated")
parser.add_argument("--subdir", type=str, default=None,
    help="sub directory to validate")
parser.add_argument("--root", type=str, required=True,
    help ="root directory with results")
parser.add_argument("--json", type=str, required=True,
    help="validation json file")
parser.add_argument("--baseline-dir", type=str, default=None,
    help="directory for baselines for regression tests, if any")
parser.add_argument("--temp-dir", type=str, default="/tmp",
    help="directory for results for regression test failures, if any")

args = parser.parse_args()

with open(args.json, 'r') as f:
    data = json.load(f)[0]

node = data
test_dir = args.root
if args.subdir is not None:
    node = find_directory(args.subdir, node)
    test_dir = os.path.join(args.root, args.subdir)

# Check that directory exists since `os.walk()` does not fail
# for non extant directories.
if not os.path.isdir(test_dir):
    raise RuntimeError("Missing directory: '%s'" % test_dir)

# Validate directory structure.
regressions_failure_count = 0
for dirpath, dirs, files in os.walk(test_dir):
    # skip hidden files and directories
    # ref: https://stackoverflow.com/questions/13454164/os-walk-without-hidden-folders
    files = [f for f in files if not f[0] == '.']
    dirs[:] = [d for d in dirs if not d[0] == '.']

    relpath = os.path.relpath(dirpath, test_dir)
    currentnode = find_directory(relpath, node)
    if not currentnode:
        raise RuntimeError("Unexpected directory was found: '%s'" % dirpath)

    dirs = set(dirs)
    files = set(files)
    expected_dirs = set()
    expected_files = set()

    regression_test = []
    for item in currentnode["contents"]:
        if item["type"] == "directory":
            expected_dirs.add(item["name"])
        elif item["type"] == "file":
            expected_files.add(item["name"])
            if item.get("compare", False):
                regression_test.append(item["name"])

    if expected_dirs != dirs:
        missing = expected_dirs - dirs
        unexpected = dirs - expected_dirs
        raise RuntimeError(\
            "Mismatched directories under '%s' found.\n"\
            "Missing : %s\n" \
            "Unexpected: %s" % (dirpath, missing, unexpected))

    if expected_files != files:
        missing = expected_files - files
        unexpected = files - expected_files
        raise RuntimeError(\
            "Mismatched files under '%s' found.\n"\
            "Missing : %s\n" \
            "Unexpected: %s" % (dirpath, missing, unexpected))

    if regression_test:
        from vtkmodules.vtkTestingRendering import vtkTesting
        t = vtkTesting()
        for image in regression_test:
            t.CleanArguments()
            t.AddArgument("-V")
            baseline_name = get_baseline(args.name, dirpath, image, args.root)
            t.AddArgument(os.path.join(args.baseline_dir, baseline_name))
            t.AddArgument("-T")
            t.AddArgument(args.temp_dir)
            if t.RegressionTest(os.path.join(dirpath, image), 15) != t.PASSED:
                regressions_failure_count += 1
                print("\n")
if regressions_failure_count:
    raise RuntimeError("ERROR: %d regression tests failed!" % regressions_failure_count)



def transform(dataset, rot_center=0, tune_rot_center=True):
    """Reconstruct sinograms using the tomopy gridrec algorithm

    Typically, a data exchange file would be loaded for this
    reconstruction. This operation will attempt to perform
    flat-field correction of the raw data using the dark and
    white background data found in the data exchange file.

    This operator also requires either the tomviz/tomopy-pipeline
    docker image, or a python environment with tomopy installed.
    """

    import numpy as np
    import tomopy

    # Get the current volume as a numpy array.
    array = dataset.active_scalars

    dark = dataset.dark
    white = dataset.white
    angles = dataset.tilt_angles
    tilt_axis = dataset.tilt_axis

    # TomoPy wants the tilt axis to be zero, so ensure that is true
    if tilt_axis == 2:
        order = [2, 1, 0]
        array = np.transpose(array, order)
        if dark is not None and white is not None:
            dark = np.transpose(dark, order)
            white = np.transpose(white, order)

    if angles is not None:
        # tomopy wants radians
        theta = np.radians(angles)
    else:
        # Assume it is equally spaced between 0 and 180 degrees
        theta = tomopy.angles(array.shape[0])

    # Perform flat-field correction of raw data
    if white is not None and dark is not None:
        array = tomopy.normalize(array, white, dark, cutoff=1.4)

    if rot_center == 0:
        # Try to find it automatically
        init = array.shape[2] / 2.0
        rot_center = tomopy.find_center(array, theta, init=init, ind=0,
                                        tol=0.5)
    elif tune_rot_center:
        # Tune the center
        rot_center = tomopy.find_center(array, theta, init=rot_center, ind=0,
                                        tol=0.5)

    # Calculate -log(array)
    array = tomopy.minus_log(array)

    # Remove nan, neg, and inf values
    array = tomopy.remove_nan(array, val=0.0)
    array = tomopy.remove_neg(array, val=0.00)
    array[np.where(array == np.inf)] = 0.00

    # Perform the reconstruction
    array = tomopy.recon(array, theta, center=rot_center, algorithm='gridrec')

    # Mask each reconstructed slice with a circle.
    array = tomopy.circ_mask(array, axis=0, ratio=0.95)

    # Set the transformed array
    child = dataset.create_child_dataset()
    child.active_scalars = array

    return_values = {}
    return_values['reconstruction'] = child
    return return_values



def transform(dataset):
    """Calculate gradient magnitude of each tilt image using Sobel operator"""

    import numpy as np
    import scipy.ndimage.filters

    array = dataset.active_scalars
    array = array.astype(np.float32)

    # Transform the dataset along the third axis.
    aaSobelX = np.empty_like(array)
    # 1D X-axis sobel
    scipy.ndimage.filters.sobel(array, axis=0, output=aaSobelX)
    aaSobelY = np.empty_like(array)
    # 1D Y-axis sobel
    scipy.ndimage.filters.sobel(array, axis=1, output=aaSobelY)

    # Calculate 2D sobel for each image slice.
    result = np.hypot(aaSobelX, aaSobelY)

    # Set the result as the new scalars.
    dataset.active_scalars = result



import tomviz.operators
import tomviz.utils

import os
import numpy as np
import scipy.ndimage

PORE_PHASE = 1
MATTER_PHASE = 0


def coord_iterator(extent):
    if len(extent) == 1:
        for i in range(extent[0][0], extent[0][1]):
            yield (i, )
    elif len(extent) > 1:
        for i in range(extent[0][0], extent[0][1]):
            for c in coord_iterator(extent[1:]):
                yield (i, ) + c


def make_sphere_structure(radius, ndim):
    shape = (2 * radius + 1,) * ndim
    center = (radius,) * ndim
    rr = radius * radius
    structure = np.empty(shape=shape, dtype=np.uint8)
    extent = tuple((0, s) for s in shape)
    for coord in coord_iterator(extent):
        if sum((a - b) * (a - b) for a, b in zip(coord, center)) <= rr:
            structure[coord] = PORE_PHASE
        else:
            structure[coord] = MATTER_PHASE
    return structure


def segment_pore(volume, threshold):
    segmented = np.empty(shape=volume.shape, dtype=np.uint8)
    segmented.fill(PORE_PHASE)
    segmented[volume > threshold] = MATTER_PHASE
    return segmented


def make_distance_map(volume):
    shape = tuple(s + 2 for s in volume.shape)
    distance = np.empty(shape=shape, dtype=volume.dtype)
    distance.fill(MATTER_PHASE)
    volume_span = tuple(slice(1, s + 1) for s in volume.shape)
    distance[volume_span] = volume
    distance = scipy.ndimage.distance_transform_edt(distance)
    return distance[volume_span]


def get_dilation_slices(coord, shape, structure_shape):
    volume_slice = []
    structure_slice = []
    for c, r, s in zip(coord, shape, structure_shape):
        start = c - s // 2
        stop = start + s
        start_offset = 0
        stop_offset = 0
        if (start < 0):
            start_offset = -start
        if (stop > r):
            stop_offset = r - stop

        volume_slice.append(slice(start + start_offset, stop + stop_offset))
        structure_slice.append(slice(start_offset, s + stop_offset))

    return tuple(volume_slice), tuple(structure_slice)


def binary_dilation(a, structure):
    """
    Alternative implementation of scipy.ndimage.binary_dilation

    The scipy version becomes very with larger structuring elements,
    and it also throws memory errors.

    This version is slower for smaller structuring elements, but becomes
    faster with larger ones, and it also appears to be more stable
    """
    dil = np.zeros(shape=a.shape, dtype=np.bool)
    extent = tuple((0, s) for s in a.shape)
    for coord in coord_iterator(extent):
        if (a[coord]):
            a_slice, structure_slice = get_dilation_slices(coord, a.shape,
                                                           structure.shape)
            dil[a_slice] = structure[structure_slice] | dil[a_slice]
    return dil


def get_dilation_count(volume, distance_map, radius, dilation_fn=None):
    if dilation_fn is None:
        dilation_fn = scipy.ndimage.binary_dilation

    structure = make_sphere_structure(radius, volume.ndim)
    pore_radii = distance_map >= radius
    r_dil = dilation_fn(pore_radii, structure=structure)
    r_dil = r_dil & volume
    return r_dil, np.sum(r_dil)


class PoreSizeDistribution(tomviz.operators.CancelableOperator):
    """Continuous pore size distribution method
    https://doi.org/10.1111/j.1551-2916.2008.02736.x
    """
    def transform(self, dataset, threshold=127, radius_spacing=1,
                  save_to_file=False, output_folder=""):
        """Operator transform method

        Args:
            threshold (int): scalars >= threshold are considered matter.
                             scalars < threshold are considered pore.
            radius_spacing (int): size distribution is computed for all radii
                                  from 1 to r_max. tweak radius spacing to
                                  reduce the number of radii. For example,
                                  if radius_spacing is 3, only calculate
                                  r_s = 1, 4, 7, ... , r_max
            save_to_file (bool): save the detailed output of the operator to
                                 files.
            output_folder (str): the path to the folder where the optional
                                 output files are written to
        """
        scalars = dataset.active_scalars

        if scalars is None:
            raise RuntimeError("No scalars found!")

        if save_to_file and not os.access(output_folder, os.W_OK):
            import warnings
            save_to_file = False
            warnings.warn(
                "Unable to write to destination folder %s" % output_folder)

        self.progress.maximum = 100
        self.progress.value = 0
        self.progress.message = "Generating the pore size map..."

        total_volume = scalars.size
        cavity = segment_pore(scalars, threshold)
        distance_map = make_distance_map(cavity)

        dataset.active_scalars = distance_map

        max_r = int(np.max(distance_map))
        # Ensure max_r is always included in the radius list
        pore_radius = list(range(1, max_r, radius_spacing))
        if pore_radius[-1] != max_r:
            pore_radius.append(max_r)

        pore_volume = []

        n = len(pore_radius)

        self.progress.message = "Calculating pore size distribution..."

        for i, r in enumerate(pore_radius):
            self.progress.value = int(100 * (i + 1) / (n + 1))
            _, count = get_dilation_count(cavity, distance_map, r,
                                          binary_dilation)
            pore_volume.append(count / total_volume)

        self.progress.value = 100

        return_values = {}

        column_names = ["Pore radius", "Pore volume"]
        table_data = np.empty(shape=(n, 2))

        table_data[:, 0] = pore_radius
        table_data[:, 1] = pore_volume

        if save_to_file:
            filename = "pore_size_distribution.csv"
            np.savetxt(os.path.join(output_folder, filename), table_data,
                       delimiter=", ", header=", ".join(column_names))

        table = tomviz.utils.make_spreadsheet(column_names, table_data)
        return_values["pore_size_distribution"] = table

        return return_values



from PIL import Image, ImageDraw

class render():

    def __init__(self):
        return

    def render(self, cis, iname, lnames, cnames):
        image = cis.get_image(iname)
        imode = "RGB"

        im = Image.new(mode=imode, size=(cis.dims[0], cis.dims[1]))

        # for now, assume only variables (no shadow, depth, etc.)
        for layer in lnames: 
            l = image.get_layer(layer)
            if l:
                shape = [(0,0),(l.dims[0], l.dims[1])]
                limage = Image.new(mode=imode, size=(l.dims[0], l.dims[1]))
                lidraw = ImageDraw.Draw(limage)
                lidraw.rectangle(shape, fill="#ffff33", outline="red")
                im.paste(limage, l.offset)

        return im




r"""
Module intended to provide functions used in Catalyst scripts. This is public
API and must preserve backwards compatibility so that older versions of Catalyst
scripts continue to work as ParaView code keeps evolving. Hence, we limit the
API exposed here to bare minimum.
"""

def Options():
    """Creates and returns an options object which is used to configure Catalyst
    specific options such as output directories, live support, etc."""
    # we import simple to ensure that the active connection is created
    from paraview import servermanager, simple
    pxm = servermanager.ProxyManager()
    proxy = pxm.NewProxy("coprocessing", "CatalystOptions")
    return servermanager._getPyProxy(proxy)

def log_level():
    """Returns the Python logging level to use to log informative messages
    from this package"""
    from paraview.modules.vtkPVVTKExtensionsCore import vtkPVLogger
    from paraview.detail.loghandler import get_level
    return get_level(vtkPVLogger.GetCatalystVerbosity())



import jsonpatch
import re
import json

from jsonpointer import resolve_pointer

from ._jsonpath import (
    operator_path,
    module_path,
    datasource_path,
    pipeline_index,
    find_operator,
    find_module,
    find_datasource,
    find_removed_object_id

)

from ._schemata import (
    load_operator,
    load_module,
    load_datasource,
    dump_module,
    dump_operator,
    dump_datasource
)

from ._models import Pipeline

from ._pipeline import PipelineStateManager


def update_list(src, dst):
    same = len(src) == len(dst)
    if len(src) == len(dst):
        for (x, y) in zip(src, dst):
            same = hasattr(x, 'id') and hasattr(y, 'id') and x.id == y.id
            if not same:
                break

    # If we are dealing with a list of the same object we can just update
    if same:
        for i in range(0, len(src)):
            x = src[i]
            if isinstance(x, built_in):
                dst[i] = x
            else:
                update(x, dst[i])
    # Otherwise, clear the list and start again
    else:
        for x in dst:
            if hasattr(x, '_kill'):
                x._kill()
        dst.clear()
        for x in src:
            dst.append(x)


built_in = (str, int, float, complex, dict)


def update(src, dst):
    for attr, value in vars(src).items():
        if isinstance(value, list):
            if not hasattr(dst, attr):
                setattr(dst, attr, value)
            else:
                update_list(value, getattr(dst, attr))
        elif isinstance(value, built_in):
            setattr(dst, attr, value)
        else:
            if not hasattr(dst, attr) or getattr(dst, attr) is None:
                setattr(dst, attr, value)
            else:
                update(value, getattr(dst, attr))

    return dst


def operator_update(patch_op):
    # First get path to operator
    path = patch_op['path']
    op_path = operator_path(path)

    # Now get the current state of the operator
    op = find_operator(op_path.split('/')[1:])
    current_state = PipelineStateManager().serialize_op(path, op.id)
    current_state_dct = json.loads(current_state)

    # Apply the update
    # Adjust the path
    patch_op['path'] = path.replace(op_path, '')
    patch = jsonpatch.JsonPatch([patch_op])
    current_state_dct = patch.apply(current_state_dct)

    # Now update the operator
    PipelineStateManager().update_op(op_path, json.dumps(current_state_dct))

    return op_path


def module_update(patch_module):
    # First get path to module
    path = patch_module['path']
    mod_path = module_path(path)

    # Now get the current state of the module
    module = find_module(mod_path.split('/')[1:])
    current_state = PipelineStateManager().serialize_module(path, module.id)
    current_state_dct = json.loads(current_state)

    # Apply the update
    # Adjust the path
    patch_module['path'] = path.replace(mod_path, '')
    patch = jsonpatch.JsonPatch([patch_module])
    current_state_dct = patch.apply(current_state_dct)

    # Now update the module
    PipelineStateManager().update_module(mod_path,
                                         json.dumps(current_state_dct))

    return mod_path


def datasource_update(patch_datasource):
    # First get path to datasource
    path = patch_datasource['path']
    ds_path = datasource_path(path)

    # Now get the current state of the datasource
    datasource = find_datasource(ds_path.split('/')[1:])
    current_state = PipelineStateManager().serialize_datasource(path,
                                                                datasource.id)
    current_state_dct = json.loads(current_state)

    # Apply the update
    # Adjust the path
    patch_datasource['path'] = path.replace(ds_path, '')
    patch = jsonpatch.JsonPatch([patch_datasource])
    current_state_dct = patch.apply(current_state_dct)

    # Now update the datasource
    PipelineStateManager().update_datasource(ds_path,
                                             json.dumps(current_state_dct))

    return ds_path


def module_add(patch_module):
    if patch_module['value'] == []:
        return

    module_json = patch_module['value']
    if isinstance(module_json, list):
        module_json = module_json[0]
    path = patch_module['path']
    ds_path = datasource_path(path)
    data_source = find_datasource(ds_path.split('/')[1:])

    state = PipelineStateManager().add_module(ds_path, data_source.id,
                                              module_json['type'])
    # If we are adding our first module just find the datasource and the first
    #  module
    parts = path.split('/')[1:]
    if path.endswith('modules'):
        module = find_datasource(parts).modules[0]
        path += '/0'
    else:
        module = find_module(parts)

    module.properties.visibility = True
    updates = module._updates()
    state = jsonpatch.JsonPatch(updates).apply(json.loads(state))

    # Now update the module
    PipelineStateManager().update_module(path, json.dumps(state))

    module_state = PipelineStateManager().serialize_module(path, state['id'])
    new_module = load_module(json.loads(module_state))
    update(new_module, module)


def operator_add(patch_op):
    if patch_op['value'] == []:
        return

    op_json = patch_op['value']

    path = patch_op['path']
    ds_path = datasource_path(path)
    data_source = find_datasource(ds_path.split('/')[1:])

    for o in op_json if isinstance(op_json, list) else [op_json]:
        op_state = PipelineStateManager().add_operator(ds_path, data_source.id,
                                                       json.dumps(o))

        op = find_operator(path.split('/')[1:])
        new_op = load_operator(json.loads(op_state))
        update(new_op, op)


def datasource_add(patch_datasource):
    if patch_datasource['value'] == []:
        return

    datasource_json = patch_datasource['value']
    # if isinstance(datasource_json, list):
    #     datasource_json = datasource_json[0]
    path = patch_datasource['path']

    datasource = find_datasource(path.split('/')[1:])
    datasource_state = \
        PipelineStateManager().add_datasource(json.dumps(datasource_json))
    new_datasource = load_datasource(json.loads(datasource_state))
    update(new_datasource, datasource)


def operator_remove(patch_op):
    path = patch_op['path']
    ds_path = datasource_path(path)
    data_source = find_datasource(ds_path.split('/')[1:])

    op_id = ''
    # Are we removing them all?
    if not path.endswith('operators'):
        op_id = find_removed_object_id(path)

    PipelineStateManager().remove_operator(path, data_source.id, op_id)


def module_remove(patch_module):
    path = patch_module['path']
    ds_path = datasource_path(path)
    data_source = find_datasource(ds_path.split('/')[1:])

    module_id = ''
    # Are we removing them all?
    if not path.endswith('modules'):
        module_id = find_removed_object_id(path)

    PipelineStateManager().remove_module(path, data_source.id, module_id)


def datasource_remove(patch_datasource):
    path = patch_datasource['path']

    ds_id = ''
    # Are we removing them all?
    if not path.endswith('dataSources'):
        ds_id = find_removed_object_id(path)

    PipelineStateManager().remove_datasource(path, ds_id)


_datasource_regx = re.compile(r'.*/dataSources/\d(.*)')


def is_datasource_update(path):
    datasource_match = _datasource_regx.match(path)

    return (datasource_match and
            (is_module_update(datasource_match.group(1)) is None) and
            (is_operator_update(datasource_match.group(1)) is None))


_module_regx = re.compile(r'.*/modules/\d(.*)')


def is_module_update(path):
    module_match = _module_regx.match(path)

    return (module_match and
            (is_datasource_update(module_match.group(1)) is None))


_op_regx = re.compile(r'.*/operators/\d(.*)')


def is_operator_update(path):
    op_match = _op_regx.match(path)

    return op_match and (is_datasource_update(op_match.group(1)) is None)


_module_add_regx = re.compile(r'.*/modules/?\d*$')


def is_module_add(path):
    return _module_add_regx.match(path)


_module_prop_add_regx = re.compile(r'.*/modules/\d*/.*$')


def is_module_prop_add(path):
    return _module_prop_add_regx.match(path)


_op_add_regx = re.compile(r'.*/operators/?\d*$')


def is_operator_add(path):
    return _op_add_regx.match(path)


_datasource_add_regx = re.compile(r'.*/dataSources/?\d*$')


def is_datasource_add(path):
    return _datasource_add_regx.match(path)


_op_remove_regx = re.compile(r'.*/operators/?\d*$')


def is_operator_remove(path):
    return _op_remove_regx.match(path)


_module_remove_regx = re.compile(r'.*/modules/?\d*$')


def is_module_remove(path):
    return _module_remove_regx.match(path)


_ds_remove_regx = re.compile(r'.*/dataSources/?\d*$')


def is_datasource_remove(path):
    return _ds_remove_regx.match(path)


def diff(src, dst):
    # The semantic for Tomviz state is a little different than a simple dict.
    # You can't change a operator or module by simply patch id of a dict. If
    # an id is different then the objects should be treated as immutable. So we
    # need to post process the operations and replace any attempts to patch
    # object ids with a operation to replace the whole object.

    # Utility function to get the prefix from a path
    def _prefix(p):
        return '/'.join(p.split('/')[:-1])

    # The new patch
    patch = []

    # The set of paths the we can skip over, as we are doing a full object
    # replace
    exclude_paths = set()

    # First get the patch from jsonpatch
    original_patch = jsonpatch.JsonPatch.from_diff(src, dst)

    # Iterate once to build up list of paths we can discard as we will be
    # simple replacing the whole object. We need todo this in two iterations as
    # we can't guarantee that the replace operation for a id will appear before
    # other operations that we need to exclude.
    for op in original_patch:
        path_prefix = _prefix(op['path'])

        if op['op'] == 'replace':
            if op['path'].endswith('/id'):
                # Save the path so we know we can ignore operations that start
                # with this path.
                exclude_paths.add(path_prefix)

    # Now iterate through and update and exclude appropriate operations.
    for op in original_patch:
        path_prefix = _prefix(op['path'])
        if op['op'] == 'replace':
            # We are trying to patch an id, replace with operation that
            # replaces the whole object.
            if op['path'].endswith('/id'):

                new_object = resolve_pointer(dst, path_prefix)
                op = {
                    "op": "replace",
                    "path": path_prefix,
                    "value": new_object
                }

                patch.append(op)
                continue

        # If not in exclude path add operation
        if path_prefix not in exclude_paths:
            patch.append(op)

    return patch


def mark_modified(ops, modules):
    ops_to_mark = {}

    for path in ops:
        pindex = pipeline_index(path)
        op_index = path[-1]
        if pindex not in ops_to_mark or op_index < ops_to_mark[pindex][-1]:
            ops_to_mark[pindex] = path

    PipelineStateManager().modified(list(ops_to_mark.values()), list(modules))


def update_app(patch, ops_modified, modules_modified):
    path = patch['path']

    if is_operator_update(path):
        ops_modified.add(operator_update(patch))
    elif is_module_update(path):
        modules_modified.add(module_update(patch))
    elif is_datasource_update(path):
        pass
        # TODO: Deal with changing file Names
        # datasources_modified.add(datasource_update(o))


def add_to_app(patch):
    path = patch['path']

    if is_module_add(path):
        module_add(patch)
    elif is_module_update(path):
        module_update(patch)
    elif is_operator_add(path):
        operator_add(patch)
    elif is_operator_update(path):
        operator_update(patch)
    elif is_datasource_add(path):
        datasource_add(patch)
    elif is_datasource_update(path):
        pass


def remove_from_app(patch):
    path = patch['path']
    if is_operator_remove(path):
        operator_remove(patch)
    elif is_operator_update(path):
        operator_update(patch)
    elif is_module_remove(path):
        module_remove(patch)
    elif is_module_update(path):
        module_update(patch)
    elif is_datasource_remove(path):
        datasource_remove(patch)
    elif is_datasource_update(path):
        pass


def convert_move_app(patch, ops_modified, modules_modified):
    # Just convert to a remove and add
    state = PipelineStateManager().serialize()
    state = json.loads(state)
    value = resolve_pointer(state, patch['from'])
    remove_from_app({
        'op': 'remove',
        'path': patch['from'],
    })

    update_app({
        'op': 'add',
        'path': patch['path'],
        'value': value
    }, ops_modified, modules_modified)


def sync_state_to_app(src, dst):
    patch = diff(src, dst)

    ops_modified = set()
    modules_modified = set()

    PipelineStateManager().disable_sync_to_python()

    for o in patch:
        patch_op = o['op']
        if patch_op == 'replace':
            update_app(o, ops_modified, modules_modified)
        elif patch_op == 'add':
            add_to_app(o)
        elif patch_op == 'remove':
            remove_from_app(o)
        elif patch_op == 'move':
            convert_move_app(o, ops_modified, modules_modified)
        else:
            raise Exception('Unexcepted op type: %s' % o['op'])

    PipelineStateManager().enable_sync_to_python()

    # Sync from app to ensure python as the updated state
    sync_state_to_python()

    mark_modified(ops_modified, modules_modified)


#
# Function for syncing state to Python
#


def operator_update_python(patch_op, removed_cache):
    # First get path to operator
    path = patch_op['path']
    op_path = operator_path(path)

    # Now get the current state of the operator
    op = find_operator(op_path.split('/')[1:])
    current_op_state = dump_operator(op)

    # Apply the update
    # Adjust the path
    patch_op['path'] = path.replace(op_path, '')
    patch = jsonpatch.JsonPatch([patch_op])
    current_op_state = patch.apply(current_op_state)
    new_op = load_operator(current_op_state, removed_cache)

    # Finally update the object inplace
    update(new_op, op)


def add_mod_to_removed_cache(removed_cache, mod):
    removed_cache['modules'][mod.id] = mod


def add_ds_to_removed_cache(removed_cache, ds):
    removed_cache['dataSources'][ds.id] = ds

    for op in ds.operators:
        add_op_to_removed_cache(removed_cache, op)

    for mod in ds.modules:
        add_mod_to_removed_cache(removed_cache, mod)


def add_op_to_removed_cache(removed_cache, op):
    removed_cache['operators'][op.id] = op

    if hasattr(op, 'dataSources'):
        for ds in op.dataSources:
            add_ds_to_removed_cache(removed_cache, ds)


def operator_remove_from_python(patch_op, removed_cache):
    # First get path to operator
    path = patch_op['path']
    op_path = operator_path(path)

    # Find parent data source
    ds = find_datasource(op_path.split('/')[1:])

    # Now get the the operator
    op = find_operator(op_path.split('/')[1:])

    # Remove from data source
    ds.operators.remove(op)

    add_op_to_removed_cache(removed_cache, op)


def operator_add_to_python(patch_op, removed_cache):
    # First get path to operator
    path = patch_op['path']
    op_path = operator_path(path)

    # Find parent data source
    ds = find_datasource(op_path.split('/')[1:])

    # Load the new ops
    value = patch_op['value']
    ops = value if isinstance(value, list) else [value]
    for op in ops:
        op = load_operator(op, removed_cache)
        ds.operators.append(op)


def module_update_python(patch_module):
    # First get path to module
    path = patch_module['path']
    mod_path = module_path(path)

    # Now get the current state of the module
    module = find_module(mod_path.split('/')[1:])
    current_module_state = dump_module(module)

    # Apply the update
    # Adjust the path
    patch_module['path'] = path.replace(mod_path, '')
    patch = jsonpatch.JsonPatch([patch_module])

    current_module_state = patch.apply(current_module_state)
    new_module = load_module(current_module_state)

    # Finally update the object inplace
    update(new_module, module)


def module_remove_from_python(patch_op, removed_cache):
    # First get path to module
    path = patch_op['path']
    mod_path = module_path(path)

    # Find parent data source
    ds = find_datasource(mod_path.split('/')[1:])

    # Now get the the module
    mod = find_module(mod_path.split('/')[1:])

    # Remove from data source
    ds.modules.remove(mod)

    # Add to cache, in case we need to resurrect it
    add_mod_to_removed_cache(removed_cache, mod)

    # Kill it
    # mod._kill()


def module_add_to_python(patch_module, removed_cache):
    # First get path to module
    path = patch_module['path']
    mod_path = module_path(path)

    # Find parent data source
    ds = find_datasource(mod_path.split('/')[1:])

    # Load the new modules
    value = patch_module['value']
    mods = value if isinstance(value, list) else [value]
    for mod in mods:
        mod = load_module(mod, removed_cache)
        ds.modules.append(mod)


def datasource_update_python(patch_ds, removed_cache):
    # First get path to module
    path = patch_ds['path']
    ds_path = datasource_path(path)

    # Now get the current state of the datasource
    ds = find_datasource(ds_path.split('/')[1:])
    current_ds_state = dump_datasource(ds)

    # Apply the update
    # Adjust the path
    patch_ds['path'] = path.replace(ds_path, '')
    patch = jsonpatch.JsonPatch([patch_ds])

    current_ds_state = patch.apply(current_ds_state)
    new_ds = load_datasource(current_ds_state, removed_cache)

    # Finally update the object inplace
    update(new_ds, ds)


def datasource_remove_from_python(patch_module, removed_cache):
    from . import pipelines

    # First get path to datasource
    path = patch_module['path']
    ds_path = datasource_path(path)

    # Find parent data source
    ds = find_datasource(ds_path.split('/')[1:])

    # Remove from pipelines
    pipeline = None
    for p in pipelines:
        if p.dataSource == ds:
            pipeline = p
            break

    if pipeline is None:
        raise Exception('Unable to find data source.')

    pipelines.remove(pipeline)

    # Kill them!
    pipeline._kill()

    add_ds_to_removed_cache(removed_cache, ds)


def datasource_add_to_python(patch_datasource, removed_cache):
    from . import pipelines
    # Load the new datasources
    value = patch_datasource['value']
    data_sources = value if isinstance(value, list) else [value]

    for ds in data_sources:
        ds = load_datasource(ds, removed_cache)
        pipelines.append(Pipeline(ds))


def add_to_python(patch, remove_cache):
    path = patch['path']

    if is_module_add(path):
        module_add_to_python(patch, remove_cache)
    elif is_module_update(path):
        module_update_python(patch)
    elif is_operator_add(path):
        operator_add_to_python(patch, remove_cache)
    elif is_operator_update(path):
        operator_update_python(patch, remove_cache)
    elif is_datasource_add(path):
        datasource_add_to_python(patch, remove_cache)
    elif is_datasource_update(path):
        pass


def update_python(patch, remove_cache):
    path = patch['path']
    if is_operator_update(path):
        operator_update_python(patch, remove_cache)
    if is_module_update(path):
        module_update_python(patch)
    elif is_datasource_update(path):
        pass


def remove_from_python(patch, removed_cache):
    path = patch['path']
    if is_module_remove(path):
        module_remove_from_python(patch, removed_cache)
    elif is_module_update(path):
        module_update_python(patch)
    elif is_operator_remove(path):
        operator_remove_from_python(patch, removed_cache)
    elif is_operator_update(path):
        operator_update_python(patch, removed_cache)
    elif is_datasource_remove(path):
        datasource_remove_from_python(patch, removed_cache)
    elif is_datasource_update(path):
        datasource_update_python(patch, removed_cache)


def convert_move_python(patch, removed_cache):
    from . import _current_state
    state = _current_state()
    # Just convert to a remove and add
    value = resolve_pointer(state, patch['from'])
    remove_from_python({
        'op': 'remove',
        'path': patch['from'],
    }, removed_cache)

    update_python({
        'op': 'add',
        'path': patch['path'],
        'value': value
    }, removed_cache)


def sync_state_to_python(current_python_state=None, current_app_state=None):
    from . import _current_state

    if current_python_state is None:
        current_python_state = _current_state()

    if current_app_state is None:
        current_app_state = PipelineStateManager().serialize()
        current_app_state = json.loads(current_app_state)

    patch = diff(current_python_state, current_app_state)

    removed_cache = {
        'modules': {},
        'operators': {},
        'dataSources': {}
    }

    for o in patch:
        patch_op = o['op']
        if patch_op == 'replace':
            update_python(o, removed_cache)
        elif patch_op == 'add':
            add_to_python(o, removed_cache)
        elif patch_op == 'remove':
            remove_from_python(o, removed_cache)
        elif patch_op == 'move':
            convert_move_python(o, removed_cache)
        else:
            raise Exception('Unexpected op type: %s' % o['op'])

    # Now kill off any removed objects
    for m in removed_cache['modules'].values():
        m._kill()

    for o in removed_cache['operators'].values():
        o._kill()

    for d in removed_cache['dataSources'].values():
        d._kill()



def generate_dataset(array, p_in=30.0, p_s=60.0, sparsity=0.20):
    import numpy as np

    arrayShape = array.shape
    x = np.fft.fftfreq(arrayShape[0])
    y = np.fft.fftfreq(arrayShape[1])
    z = np.fft.fftfreq(arrayShape[2])

    X, Y, Z = np.meshgrid(y, x, z)
    kr = np.sqrt(X**2 + Y**2 + Z**2)

    # Create inner structure.
    A = np.exp(-p_in * kr) # Generate amplitude
    phase = np.random.randn(arrayShape[0], arrayShape[
                            1], arrayShape[2]) # Generate phase
    F = A * np.exp(2 * np.pi * 1j * phase) # Combine amplitude and phase
    f = np.fft.ifftn(F) # Inverse FFT
    f_in = np.absolute(f).copy().flatten()

    # Create shape
    A = np.exp(-p_s * kr) # Generate amplitude
    phase = np.random.randn(arrayShape[0], arrayShape[
                            1], arrayShape[2]) # Generate phase
    F = A * np.exp(2 * np.pi * 1j * phase) # Combine amplitude and phase
    f = np.fft.ifftn(F) # Inverse FFT
    f_shape = np.absolute(f)

    # Impose sparsity (% of non-zero voxels)
    f_shape = np.argsort(f_shape, axis=None) # Sort the shape image
    f_shape = f_shape.flatten()
    # Number of zero voxels
    N_zero = np.int(np.round((array.size * (1 - sparsity))))
    f_shape[N_zero:] = f_shape[N_zero]
    f_in[f_shape] = 0

    f_in = f_in / np.amax(f_in) # Normalize image
    np.copyto(array, f_in.reshape(arrayShape)) # Update array
